"""Builtin float point values."""

__all__ = ["FloatDataType", "Float", "FloatType"]

from functools import partialmethod

from acaciamc.error import Error, ErrorType
from acaciamc.tools import axe, cmethod_of
from acaciamc.mccmdgen.datatype import DefaultDataType
from acaciamc.mccmdgen.ctexpr import CTDataType
from acaciamc.mccmdgen.expr import *
from acaciamc.mccmdgen.utils import InvalidOpError
from .integer import IntLiteral
from .types import Type

class FloatDataType(DefaultDataType):
    name = "float"

ctdt_float = CTDataType("float")

class FloatType(Type):
    def do_init(self):
        @cmethod_of(self, "__new__")
        @axe.chop
        @axe.arg("i", axe.LiteralInt())
        @axe.slash
        def _new(compiler, i: int):
            return Float(float(i))

    def datatype_hook(self):
        return FloatDataType()

    def cdatatype_hook(self):
        return ctdt_float

class Float(ConstExprCombined):
    cdata_type = ctdt_float

    def __init__(self, value: float):
        super().__init__(FloatDataType())
        self.value = value

    def cstringify(self) -> str:
        return str(self)

    def chash(self):
        return self.value

    @classmethod
    def from_int(cls, integer: IntLiteral):
        return Float(float(integer.value))

    def __str__(self) -> str:
        return str(self.value)

    def cunarypos(self):
        return self

    def cunaryneg(self):
        return Float(-self.value)

    def _bin_op(self, other, method: str):
        """`method`: "__add__", "__sub__", etc."""
        if isinstance(other, (Float, IntLiteral)):
            try:
                v = getattr(self.value, method)(other.value)
            except ArithmeticError as err:
                raise Error(ErrorType.CONST_ARITHMETIC, message=str(err))
            return Float(v)
        raise InvalidOpError

    cadd = partialmethod(_bin_op, method="__add__")
    csub = partialmethod(_bin_op, method="__sub__")
    cmul = partialmethod(_bin_op, method="__mul__")
    cdiv = partialmethod(_bin_op, method="__truediv__")
    cmod = partialmethod(_bin_op, method="__mod__")
    cradd = partialmethod(_bin_op, method="__radd__")
    crsub = partialmethod(_bin_op, method="__rsub__")
    crmul = partialmethod(_bin_op, method="__rmul__")
    crdiv = partialmethod(_bin_op, method="__rtruediv__")
    crmod = partialmethod(_bin_op, method="__rmod__")



"""Command abstraction optimizer."""

__all__ = ["Optimizer"]

from typing import Iterable, Dict, Set, List, Tuple
from abc import ABCMeta, abstractmethod

import acaciamc.mccmdgen.cmds as cmds
from acaciamc.mccmdgen.utils import unreachable

class Optimizer(cmds.FunctionsManager, metaclass=ABCMeta):
    def optimize(self):
        """Start optimizing."""
        self.opt_empty_functions()
        self.opt_dead_functions()
        self.opt_execute_as_ats()
        self.opt_function_inliner()

    @abstractmethod
    def entry_files(self) -> Iterable[cmds.MCFunctionFile]:
        """Must be implemented by subclasses to mark mcfunction files
        as an entry so that optimizer won't delete them even if they
        are not referenced.
        """
        pass

    def opt_dead_functions(self):
        """Remove unreferenced functions."""
        ref_map: Dict[cmds.MCFunctionFile, Set[cmds.MCFunctionFile]] = {}
        for file in self.files:
            refs = set(command.func_ref() for command in file.commands)
            if None in refs:
                refs.remove(None)
            ref_map[file] = refs
        visited = set()
        def _visit(file: cmds.MCFunctionFile):
            if file in visited:
                return
            visited.add(file)
            for ref in ref_map[file]:
                _visit(ref)
        for file in self.entry_files():
            _visit(file)
        self.files = list(visited)

    @property
    @abstractmethod
    def max_inline_file_size(self) -> int:
        pass

    def dont_inline_execute_call(self, file: cmds.MCFunctionFile) -> bool:
        """When True is returned, `execute ... run function` in `file`
        will not be inlined by `opt_function_inliner`.
        """
        return False

    def _resolve_execute(self, c: cmds.Command) \
            -> Tuple[List[cmds._ExecuteSubcmd], cmds.Command]:
        subcmds = []
        while isinstance(c, cmds.Execute):
            subcmds.extend(c.subcmds)
            c = c.runs
        return subcmds, c

    def opt_function_inliner(self):
        """Expand mcfunctions that only have 1 reference."""
        # Locating optimizable
        todo: Dict[cmds.MCFunctionFile,
                   Tuple[cmds.MCFunctionFile, int, bool]] = {}
        called: Set[cmds.MCFunctionFile] = set()
        for file in self.files:
            allow_execute = not self.dont_inline_execute_call(file)
            for i, command in enumerate(file.commands):
                callee = command.func_ref()
                if callee is None:
                    continue
                if callee in called:
                    # More than 1 references
                    if callee in todo:
                        del todo[callee]
                    continue
                called.add(callee)
                subcmds, runs = self._resolve_execute(command)
                if not isinstance(runs, cmds.InvokeFunction):
                    # Not a direct /function call
                    continue
                passed_test = (
                    (not subcmds)
                    or (
                        # If there is /execute...
                        # Caller allows inlining calls with execute
                        allow_execute
                        # Environments other than if/unless may change
                        # during execution of commands, so we can't
                        # inline it.
                        and (all([
                            isinstance(
                                subcmd, (
                                    cmds.ExecuteScoreComp,
                                    cmds.ExecuteScoreMatch,
                                    cmds.ExecuteCond
                                )
                            )
                            for subcmd in subcmds
                        ]))
                    )
                )
                if passed_test or callee.cmd_length() == 1:
                    todo[callee] = (file, i, not passed_test)
        # Optimize
        # for callee, (caller, caller_index) in todo.items():
        #     print("Inlining %s, called by %s at %d" % (
        #         callee.get_path(), caller.get_path(), caller_index
        #     ))
        merged: Set[cmds.MCFunctionFile] = set()
        def _need_tmp(subcmds, commands: List[cmds.Command]) -> bool:
            slots = set()
            for subcmd in subcmds:
                if isinstance(subcmd, cmds.ExecuteCond):
                    return True
                if isinstance(subcmd, cmds.ExecuteScoreComp):
                    slots.add(subcmd.operand1)
                    slots.add(subcmd.operand2)
                elif isinstance(subcmd, cmds.ExecuteScoreMatch):
                    slots.add(subcmd.operand)
                else:
                    unreachable()
            for command in commands:
                for slot in slots:
                    if command.scb_did_assign(slot):
                        return True
            return False
        def _merge(caller: cmds.MCFunctionFile):
            if caller in merged:
                return
            merged.add(caller)
            tasks: List[Tuple[int, cmds.MCFunctionFile, bool]] = []
            for callee, (caller_got, caller_index, ensure_l1) in todo.items():
                if caller_got != caller:
                    continue
                _merge(callee)
                tasks.append((caller_index, callee, ensure_l1))
            tasks.sort(key=lambda x: x[0], reverse=True)
            for index, callee, ensure_len1 in tasks:
                subcmds, _ = self._resolve_execute(caller.commands[index])
                callee_len = callee.cmd_length()
                # Here `runs` must be /function that calls `callee`
                if subcmds and callee_len > self.max_inline_file_size:
                    # Prefixing every command in a long file with
                    # /execute condition can reduce performance,
                    # so we don't inline it.
                    continue
                if ensure_len1 and callee_len != 1:
                    continue
                # print("Merging %s to %s:%d" % (
                #     callee.get_path(), caller.get_path(), index
                # ))
                inserts = []
                # No need for tmp if callee has only 1 command
                if callee_len != 1 and _need_tmp(subcmds, callee.commands):
                    tmp = self.allocate()
                    inserts.append(cmds.ScbSetConst(tmp, 0))
                    inserts.append(cmds.Execute(
                        subcmds, cmds.ScbSetConst(tmp, 1)
                    ))
                    for command in callee.commands:
                        inserts.append(cmds.execute(
                            [cmds.ExecuteScoreMatch(tmp, "1")], command
                        ))
                elif subcmds:
                    for command in callee.commands:
                        inserts.append(cmds.execute(subcmds, command))
                else:
                    inserts.extend(callee.commands)
                fp = callee.get_path()
                inserts.insert(0, cmds.Comment(
                    "## Function call to %s inlined by optimizer" % fp
                ))
                inserts.append(cmds.Comment("## Inline of %s ended" % fp))
                caller.commands[index : index+1] = inserts
                self.files.remove(callee)
        for caller, _, _ in todo.values():
            if caller not in todo:
                _merge(caller)

    def opt_execute_as_ats(self):
        """Remove "as @s" in /execute commands. """
        for file in self.files:
            for i, command in enumerate(file.commands):
                if isinstance(command, cmds.Execute):
                    for j, subcmd in enumerate(command.subcmds):
                        if (isinstance(subcmd, cmds.ExecuteEnv)
                                and subcmd.cmd == "as"
                                and subcmd.args == "@s"):
                            del command.subcmds[j]
                    if not command.subcmds:
                        # /execute with only a run subcommand can get
                        # rid of the /execute.
                        file.commands[i] = command.runs

    def opt_empty_functions(self):
        """Remove definition and invoke of empty functions."""
        removed = set()
        for i, file in reversed(tuple(enumerate(self.files))):
            if not file.has_content():
                removed.add(file)
                del self.files[i]
        for file in self.files:
            for i, command in enumerate(file.commands):
                ref = command.func_ref()
                if ref is not None and ref in removed:
                    file.commands[i] = cmds.Comment(
                        "## (Calling empty function) %s" % command.resolve()
                    )



# Visualize the Abstract Grammar Tree of code

# Add `acaciamc` directory to path
import os
import sys
sys.path.append(os.path.realpath(
    os.path.join(__file__, os.pardir, os.pardir)
))

import io
from types import MethodType
from typing import List, Dict, Type as PyType

from acaciamc.tokenizer import Tokenizer, TokenType
from acaciamc.parser import Parser
from acaciamc.error import *
from acaciamc.ast import AST

MC_VERSION = (1, 19, 80)

class ASTVisualizer:
    _fields_cache: Dict[PyType[AST], List[str]] = {}

    def __init__(self, node: AST, indent=2):
        self.node = node
        self.indent = indent
        super().__init__()

    @classmethod
    def get_fields(cls, node: AST) -> List[str]:
        """Get the names of fields of a node."""
        tp = type(node)
        if tp not in cls._fields_cache:
            cls._fields_cache[tp] = [
                name for name in dir(node)
                if (
                    not name.startswith('_')
                    and name != 'lineno'
                    and name != 'col'
                    and name != 'show_debug'
                )
            ]
        return cls._fields_cache[tp]

    def _convert(self, value, indent: int = 0) -> str:
        res: List[str] = []
        indent_next = indent + self.indent
        if isinstance(value, AST):
            res.append(
                '@%d:%d %s(\n' % (
                    value.lineno, value.col,
                    value.__class__.__name__
                )
            )
            for field in self.get_fields(value):
                fvalue = getattr(value, field)
                if isinstance(fvalue, MethodType):
                    continue
                res.append('%s%s = %s\n' % (
                    ' ' * indent_next, field,
                    self._convert(fvalue, indent=indent_next)
                ))
            res.append('%s)' % (' ' * indent))
        elif isinstance(value, list):
            res.append('[\n')
            for element in value:
                sub = self._convert(element, indent=(indent + self.indent))
                res.append('%s%s\n' % (' ' * indent_next, sub))
            res.append('%s]' % (' ' * indent))
        elif isinstance(value, dict):
            res.append('{\n')
            for k, v in value.items():
                sub = self._convert(v, indent=(indent + self.indent))
                res.append('%s%r: %s\n' % (' ' * indent_next, k, sub))
            res.append('%s}' % (' ' * indent))
        else:
            res.append(repr(value))
        return "".join(res)

    def get_string(self) -> str:
        return self._convert(self.node)

def test_tokenize(src: str):
    print('===TOKENIZER===')
    tk = Tokenizer(io.StringIO(src), MC_VERSION)
    while True:
        token = tk.get_next_token()
        print(repr(token))
        if token.type is TokenType.end_marker:
            break

def test_parser(src: str):
    print('===PARSER===')
    parser = Parser(Tokenizer(io.StringIO(src), MC_VERSION))
    visualizer = ASTVisualizer(parser.module())
    print(visualizer.get_string())

def test(src: str):
    test_tokenize(src)
    test_parser(src)

definition_test = '''
interface x:
    pass
def x(a: int, b=True) -> a:
    pass
'''

if_test = '''if a:
    # a comment

    #*pass*# pass
    if a + 1:
        1
elif a:  # a comment
    pass'''

attr_test = '''
aaa
a.b(1, a+2, c=3,)
b.c().d
'''

operator_test = '''
a > b > c + 1 <= 4 and not 11 > 6 or b
'''

assign_test = '''
"money":"@p" = 1
b -> "money":"@p"
a.b.c *= 5 % b
'''

backslash_test = '''
a = \\
    3
    \\
  # a comment

x = \\
  10 \\
    + 20

'''

entity_test = '''
entity X:
    x: int
    inline def get() -> int:
        result self.x

entity Y extends X:
    def get():
        result upcast(self, X).get() + 1
'''

decimal_test = '''
1.2  # 1.2
02.foo  # (2).foo
0x10.f  # (16).f (float does not support hex form)
'''

multiline_test = '''
 #* mul *# if x:
    #* xx
    g \\
        y
        *# (y
      # xxx
    + z)
        /*sss ${1}
  sss*/ 5'''

list_map_test = '''
{1, 2, 3}
{1, 2,}
{1: 2, 3: 4}
{{1: 2,}, 3, {4}}
{:}
{}
'''

command_test = '''
/sss
/*  a \${ */
/s${xxx}ss
{}
/*${1
}
f
*/
/*xxx${aaa *
    some_expr +
    # comment
    1 \\
        + {"embedded"}
}*/
/*${"xx\${xx${{xxx}[0] + "qqq${1}ppp"}ggg"}*/
'''

new_test = '''
entity X:
    new(x: int):
        new(xxx)
entity Y extends X:
    inline new(&y: bool):
        X.new(yyy)
'''

try:
    test(new_test)
except Error as err:
    err.location.file = "<testsrc>"
    print("Error:", err)



"""Abstarct Syntax Tree definitions for Acacia."""

from typing import (
    Union as _Union, List as _List, Optional as _Optional, Dict as _Dict,
    Iterable as _Iterable
)

from acaciamc.localization import LocalizedEnum as _LocalizedEnum

####################
### AST CONTENTS ###
####################

class Operator(_LocalizedEnum):
    # unary
    positive = "ast.operator.positive"
    negative = "ast.operator.negative"
    not_ = "ast.operator.not"
    # binary
    multiply = "ast.operator.multiply"
    divide = "ast.operator.divide"
    mod = "ast.operator.mod"
    add = "ast.operator.add"
    minus = "ast.operator.minus"
    # compare
    equal_to = "ast.operator.equal_to"
    unequal_to = "ast.operator.unequal_to"
    greater = "ast.operator.greater"
    less = "ast.operator.less"
    greater_equal = "ast.operator.greater_equal"
    less_equal = "ast.operator.less_equal"
    # boolean
    and_ = "ast.operator.and"
    or_ = "ast.operator.or"

COMPOP_INVERT = {
    # Used to invert ("not") a comparison
    Operator.greater: Operator.less_equal,
    Operator.greater_equal: Operator.less,
    Operator.less: Operator.greater_equal,
    Operator.less_equal: Operator.greater,
    Operator.equal_to: Operator.unequal_to,
    Operator.unequal_to: Operator.equal_to
}
COMPOP_SWAP = {
    # Used to swap the operands of a comparison
    Operator.greater: Operator.less,
    Operator.greater_equal: Operator.less_equal,
    Operator.less: Operator.greater,
    Operator.less_equal: Operator.greater_equal,
    Operator.equal_to: Operator.equal_to,
    Operator.unequal_to: Operator.unequal_to
}

class MethodQualifier(_LocalizedEnum):
    """Entity method qualifiers."""

    none = "ast.methodqualifier.none"
    virtual = "ast.methodqualifier.virtual"
    override = "ast.methodqualifier.override"
    static = "ast.methodqualifier.static"

class FuncPortType(_LocalizedEnum):
    """Function port types."""

    by_value = "ast.funcporttype.by_value"
    by_reference = "ast.funcporttype.by_reference"
    const = "ast.funcporttype.const"

class ModuleMeta:
    """Specifies a module."""
    def __init__(self, last_name: str, leading_dots: int = 0,
                 parents: _Iterable[str] = ()):
        self.leading_dots = leading_dots
        self.last_name = last_name
        self.parents = list(parents)

    def __str__(self) -> str:
        return ("." * self.leading_dots
                + ".".join(self.parents)
                + ("." if self.parents else "")
                + self.last_name)

    def __repr__(self) -> str:
        return "<ModuleMeta %r>" % str(self)

#################
### AST NODES ###
#################

class AST:

    show_debug = True  # show debug info when visited

    def __init__(self, lineno: int, col: int):
        # lineno & col: position where this node starts
        self.lineno = lineno
        self.col = col

# these classes are for classifying

class Statement(AST):
    pass

class Expression(AST):
    show_debug = False

# details

class Module(AST):  # a module
    def __init__(self, body: _List[Statement], lineno, col):
        super().__init__(lineno, col)
        self.body = body

class ArgumentTable(AST):  # arguments used in function definition
    show_debug = False

    def __init__(self, lineno, col):
        super().__init__(lineno, col)
        self.args: _List[str] = []  # argument names
        self.default: _Dict[str, _Optional[Expression]] = {}
        self.types: _Dict[str, FunctionPort] = {}

    def add_arg(self, name: str, type_: "FunctionPort",
                default: _Optional[Expression]):
        self.args.append(name)
        self.types[name] = type_
        self.default[name] = default

class CallTable(AST):  # call table
    show_debug = False

    def __init__(self, args: _List[Expression],
                 keywords: _Dict[str, Expression], lineno, col):
        super().__init__(lineno, col)
        self.args = args
        self.keywords = keywords

class TypeSpec(AST):  # specify type of value `int`
    show_debug = False

    def __init__(self, content: Expression, lineno, col):
        super().__init__(lineno, col)
        self.content = content

class FunctionPort(AST):
    show_debug = False

    def __init__(self, type_: _Optional[TypeSpec],
                 port: FuncPortType, lineno, col):
        super().__init__(lineno, col)
        self.type = type_
        self.port = port

class FormattedStr(AST):  # a literal string with ${formatted exprs}
    show_debug = False

    def __init__(self, content: _List[_Union[Expression, str]], lineno, col):
        super().__init__(lineno, col)
        self.content = content

class ExprStatement(Statement):  # a statement that is an expression
    def __init__(self, value: Expression, lineno, col):
        super().__init__(lineno, col)
        self.value = value

class Pass(Statement):  # does nothing
    pass

class If(Statement):  # if statement
    def __init__(
        self, condition: Expression,
        body: _List[Statement], else_body: _List[Statement], lineno, col
    ):
        super().__init__(lineno, col)
        self.condition = condition
        self.body = body
        self.else_body = else_body

class While(Statement):  # while statement
    def __init__(self, condition: Expression,
                 body: _List[Statement], lineno, col):
        super().__init__(lineno, col)
        self.condition = condition
        self.body = body

class FuncData(AST):
    show_debug = False

class FuncDef(Statement):  # function definition
    def __init__(self, name: str, data: FuncData, lineno, col):
        super().__init__(lineno, col)
        self.name = name
        self.data = data

class NormalFuncData(FuncData):
    def __init__(
        self, arg_table: ArgumentTable,
        body: _List[Statement], returns: _Optional[FunctionPort], lineno, col
    ):
        super().__init__(lineno, col)
        self.arg_table = arg_table
        if returns is None:
            self.returns = None
        else:
            assert returns.port is FuncPortType.by_value
            assert returns.type is not None
            self.returns = returns.type
        self.body = body

class InlineFuncData(FuncData):
    def __init__(
        self, arg_table: ArgumentTable,
        body: _List[Statement], returns: _Optional[FunctionPort], lineno, col
    ):
        super().__init__(lineno, col)
        self.arg_table = arg_table
        self.returns = returns
        self.body = body

class ConstFuncData(FuncData):
    def __init__(
        self, arg_table: ArgumentTable,
        body: _List[Statement], returns: _Optional[FunctionPort], lineno, col
    ):
        super().__init__(lineno, col)
        self.arg_table = arg_table
        self.returns = returns
        self.body = body
        assert returns is None or returns.port is not FuncPortType.const

class InterfaceDef(Statement):  # define an interface
    def __init__(self, path: _Union[str, "StrLiteral"],
                 body: _List[Statement], lineno, col):
        super().__init__(lineno, col)
        self.path = path
        self.body = body

class EntityField(Statement):  # entity field definition
    def __init__(self, name: str, type_: TypeSpec, lineno, col):
        super().__init__(lineno, col)
        self.name = name
        self.type = type_

class EntityMethod(Statement):  # entity method definition
    def __init__(self, content: FuncDef,
                 qualifier: MethodQualifier, lineno, col):
        super().__init__(lineno, col)
        self.content = content
        self.qualifier = qualifier
        assert (not isinstance(content.data, ConstFuncData)
                or qualifier is MethodQualifier.static)

class NewMethod(Statement):  # new method definition
    def __init__(self, data: _Union[NormalFuncData, InlineFuncData],
                 lineno, col):
        super().__init__(lineno, col)
        self.data = data

class EntityTemplateDef(Statement):  # entity statement
    def __init__(
        self, name: str, parents: _List[Expression],
        body: _List[_Union[EntityMethod, EntityField, Pass]],
        new_method: _Optional[NewMethod],
        lineno, col
    ):
        super().__init__(lineno, col)
        self.name = name
        self.parents = parents
        self.body = body
        self.new_method = new_method

class VarDef(Statement):  # x: y [= z] variable declaration
    def __init__(self, target: str, type_: TypeSpec,
                 value: _Optional[Expression], lineno, col):
        super().__init__(lineno, col)
        self.target = target
        self.type = type_
        self.value = value

class AutoVarDef(Statement):  # := short variable declaration
    def __init__(self, target: str, value: Expression, lineno, col):
        super().__init__(lineno, col)
        self.target = target
        self.value = value

class Assign(Statement):  # normal assign
    def __init__(self, target: Expression, value: Expression, lineno, col):
        super().__init__(lineno, col)
        self.target = target
        self.value = value

class ConstDef(Statement):  # constant definition
    def __init__(self, names: _List[str], types: _List[_Optional[TypeSpec]],
                 values: _List[Expression], lineno, col):
        super().__init__(lineno, col)
        self.names = names
        self.types = types
        self.values = values

class ReferenceDef(Statement):  # reference definition
    def __init__(self, name: str, type_: _Optional[TypeSpec],
                 value: Expression, lineno, col):
        super().__init__(lineno, col)
        self.name = name
        self.type = type_
        self.value = value

class Command(Statement):  # raw command
    def __init__(self, content: FormattedStr, lineno, col):
        super().__init__(lineno, col)
        self.content = content

class AugmentedAssign(Statement):  # augmented assign
    def __init__(
        self, target: Expression, operator: Operator,
        value: Expression, lineno, col
    ):
        super().__init__(lineno, col)
        self.target = target
        self.operator = operator
        self.value = value

class Import(Statement):  # import a module
    def __init__(self, meta: ModuleMeta, alias: _Optional[str], lineno, col):
        super().__init__(lineno, col)
        self.meta = meta
        self.name = self.meta.last_name if alias is None else alias

class FromImport(Statement):  # import specific things from a module
    def __init__(
        self, meta: ModuleMeta, names: _List[str],
        aliases: _List[_Optional[str]], lineno, col
    ):
        super().__init__(lineno, col)
        self.meta = meta
        self.id2name: _Dict[str, str] = {}
        for name, alias in zip(names, aliases):
            self.id2name[name] = name if alias is None else alias

class FromImportAll(Statement):  # import everything in a module
    def __init__(self, meta: ModuleMeta, lineno, col):
        super().__init__(lineno, col)
        self.meta = meta

class For(Statement):  # for-in iteration
    def __init__(self, name: str, expr: Expression,
                 body: _List[Statement], lineno, col):
        super().__init__(lineno, col)
        self.name = name
        self.expr = expr
        self.body = body

class StructField(Statement):  # a struct's field
    def __init__(self, name: str, type_: TypeSpec, lineno, col):
        super().__init__(lineno, col)
        self.name = name
        self.type = type_

class StructDef(Statement):  # struct definition
    def __init__(self, name: str, bases: _List[Expression],
                 body: _List[_Union[StructField, Pass]], lineno, col):
        super().__init__(lineno, col)
        self.name = name
        self.bases = bases
        self.body = body

class Result(Statement):  # result xxx
    def __init__(self, value: Expression, lineno, col):
        super().__init__(lineno, col)
        self.value = value

class NewCall(Statement):  # Template.new() or new()
    def __init__(self, primary: _Optional[Expression],
                 call_table: CallTable, lineno, col):
        super().__init__(lineno, col)
        self.primary = primary
        self.call_table = call_table

class Literal(Expression):  # a literal constant
    def __init__(self, literal, lineno, col):
        super().__init__(lineno, col)
        self.value = literal

class StrLiteral(Expression):  # a string literal
    def __init__(self, content: FormattedStr, lineno, col):
        super().__init__(lineno, col)
        self.content = content

class Self(Expression):  # "self" keyword
    pass

class Identifier(Expression):  # an identifier
    def __init__(self, name: str, lineno, col):
        super().__init__(lineno, col)
        self.name = name

class UnaryOp(Expression):  # +x, -x, not x
    def __init__(self, operator: Operator, operand: Expression, lineno, col):
        super().__init__(lineno, col)
        self.operator = operator
        self.operand = operand

class BinOp(Expression):  # an expr with binary operator (+, %, >=, etc)
    def __init__(self, left: Expression, operator: Operator,
                 right: Expression, lineno, col):
        super().__init__(lineno, col)
        self.left = left
        self.operator = operator
        self.right = right

class Call(Expression):  # call a name
    def __init__(self, func: Expression, table: CallTable, lineno, col):
        super().__init__(lineno, col)
        self.func = func
        self.table = table

class Attribute(Expression):  # attr of an expr
    def __init__(self, object_: Expression, attr: str, lineno, col):
        super().__init__(lineno, col)
        self.object = object_
        self.attr = attr

class Subscript(Expression):  # value[v1, v2]
    def __init__(self, object_: Expression,
                 subscripts: _List[Expression], lineno, col):
        super().__init__(lineno, col)
        self.object = object_
        self.subscripts = subscripts

class CompareOp(Expression):  # ==, !=, >, <, >=, <=
    def __init__(
        self, left: Expression, operators: _List[Operator],
        operands: _List[Expression], lineno, col
    ):
        super().__init__(lineno, col)
        self.left = left
        self.operators = operators
        self.operands = operands

class BoolOp(Expression):  # and, or
    def __init__(self, operator: Operator,
                 operands: _List[Expression], lineno, col):
        super().__init__(lineno, col)
        self.operator = operator
        self.operands = operands

class ListDef(Expression):  # a literal compile time list
    def __init__(self, items: _List[Expression], lineno, col):
        super().__init__(lineno, col)
        self.items = items

class MapDef(Expression):  # a literal compile time map
    def __init__(self, keys: _List[Expression],
                 values: _List[Expression], lineno, col):
        super().__init__(lineno, col)
        self.keys = keys
        self.values = values

#############
### UTILS ###
#############

class ASTVisitor:
    """Base class of an AST handler."""
    def visit(self, node: AST, **kwargs):
        visitor = getattr(
            self,
            'visit_%s' % node.__class__.__name__,
            self.general_visit
        )
        return visitor(node, **kwargs)

    def general_visit(self, node: AST):
        raise NotImplementedError



from typing import Union, TYPE_CHECKING

from acaciamc.objects import *
from acaciamc.mccmdgen.expr import *
from acaciamc.mccmdgen.datatype import DataType
from acaciamc.tools import axe, resultlib
from acaciamc.error import Error, ErrorType
from acaciamc.localization import localize
from acaciamc.mccmdgen.ctexpr import CTDataType
import acaciamc.mccmdgen.cmds as cmds

if TYPE_CHECKING:
    from acaciamc.compiler import Compiler
    from acaciamc.objects.entity import _EntityBase

class AnyDataType(DataType):
    def __str__(self) -> str:
        return 'Any'

    @classmethod
    def name_no_generic(cls) -> str:
        return 'Any'

    def matches(self, other: "DataType") -> bool:
        return True

class _AnyCTDataType(CTDataType):
    def is_baseof(self, other: CTDataType) -> bool:
        return True

any_ctdt = _AnyCTDataType("Any")

class AnyType(Type):
    def datatype_hook(self):
        return AnyDataType()

    def cdatatype_hook(self):
        return any_ctdt

@axe.chop
@axe.arg("x", axe.AnyValue())
@axe.arg("y", axe.AnyValue())
@axe.slash
def swap(compiler: "Compiler", x: AcaciaExpr, y: AcaciaExpr):
    """
    swap(&x, &y, /)
    Swap values of two variables.
    """
    if not x.is_assignable():
        raise axe.ArgumentError(
            "x", localize("modules.builtins.swap.assignable")
        )
    if not y.is_assignable():
        raise axe.ArgumentError(
            "y", localize("modules.builtins.swap.assignable")
        )
    if not (x.data_type.is_type_of(y) and y.data_type.is_type_of(x)):
        raise axe.ArgumentError(
            "x",
            localize("modules.builtins.swap.differenttype")
            % (x.data_type, y.data_type)
        )
    return resultlib.commands(swap_exprs(x, y, compiler))

@axe.chop
@axe.arg("target", axe.AnyOf(axe.LiteralString(), axe.Typed(EntityDataType)))
@axe.arg("objective", axe.LiteralString())
def scb(compiler, target: Union["_EntityBase", str], objective: str):
    """
    scb(target: str | entity, objective: str) -> &int
    Return a reference to the score of `target` on `objective`. This is
    used to interact with scores on scoreboard.
    """
    tg = target if isinstance(target, str) else target.to_str()
    return IntVar(cmds.ScbSlot(tg, objective))

@axe.chop
@axe.arg("object", EntityDataType, rename="obj")
@axe.arg("template", ETemplateDataType)
@axe.slash
def upcast(compiler, obj: "_EntityBase", template: EntityTemplate):
    """
    upcast(object: entity, template: entity_template) -> entity
    Up-cast entity `object` to `template`.
    """
    if not obj.template.is_subtemplate_of(template):
        raise Error(ErrorType.INVALID_UPCAST,
                    t1=obj.template.name, t2=template.name)
    return obj.cast_to(template)

def acacia_build(compiler: "Compiler"):
    res = {}
    # builtin types
    for name, cls in (
        ('int', IntType),
        ('bool', BoolType),
        ('Pos', PosType),
        ('Rot', RotType),
        ('Offset', PosOffsetType),
        ('Engroup', EGroupGeneric),
        ('Enfilter', EFilterType),
        ('list', ListType),
        ('map', MapType),
        ('AbsPos', AbsPosType),
        ('Any', AnyType),
        ('float', FloatType),
    ):
        res[name] = cls()
    res['ExternEngroup'] = ExternEGroupType(compiler)
    # builtin names
    res['Entity'] = compiler.base_template
    res['swap'] = BinaryFunction(swap)
    res['scb'] = BinaryFunction(scb)
    res['upcast'] = BinaryFunction(upcast)
    return res



"""
Compile-time only mapping objects.
Only a few literal values are supported as keys.
"""

__all__ = ["MapType", "MapDataType", "Map", "CTConstMap", "CTMap"]

from typing import Dict, Hashable, Iterable, Optional, Tuple, List
from itertools import repeat

from .types import Type
from .none import NoneLiteral
from .list_ import AcaciaList, CTList, list2ct
from .boolean import BoolLiteral
from .integer import IntLiteral
from acaciamc.tools import axe, cmethod_of
from acaciamc.error import *
from acaciamc.mccmdgen.expr import *
from acaciamc.mccmdgen.datatype import DefaultDataType
from acaciamc.mccmdgen.ctexpr import CTObj, CTDataType, CTObjPtr, CTExpr
from acaciamc.mccmdgen.utils import InvalidOpError

class MapDataType(DefaultDataType):
    name = "const_map"

ctdt_constmap = CTDataType("const_map")
ctdt_map = CTDataType("map", (ctdt_constmap,))

class MapType(Type):
    def do_init(self):
        @cmethod_of(self, "__new__")
        @axe.chop
        @axe.arg("iter1", axe.CTIterator())
        @axe.arg("iter2", axe.CTIterator())
        def _new(compiler, iter1: List[CTObj], iter2: List[CTObj]):
            return CTMap(iter1, iter2)
        @cmethod_of(self, "from_keys")
        @axe.chop
        @axe.arg("x", axe.CTIterator())
        @axe.slash
        @axe.arg("fill", axe.Constant(), default=NoneLiteral())
        def _from_keys(compiler, x: List[CTObj], fill: CTObj):
            return CTMap(x, repeat(fill))

    def datatype_hook(self):
        return MapDataType()

    def cdatatype_hook(self):
        return ctdt_constmap

class Map(ConstExpr):
    def __init__(self, keys: Iterable[ConstExpr],
                 values: Iterable[ConstExpr]):
        super().__init__(MapDataType())
        self.data: Dict[Hashable, Tuple[ConstExpr, ConstExpr]] = {}
        for key, value in zip(keys, values):
            self.set(key, value)
        @cmethod_of(self, "__getitem__")
        @axe.chop
        @axe.arg("key", axe.AnyValue())
        def _getitem(compiler, key: AcaciaExpr) -> ConstExpr:
            res = self.get(key)
            if res is None:
                raise Error(ErrorType.MAP_KEY_NOT_FOUND)
            return res
        @cmethod_of(self, "copy")
        @axe.chop
        def _copy(compiler):
            res = Map((), ())
            res.data = self.data.copy()
            return res
        @cmethod_of(self, "keys")
        @axe.chop
        def _keys(compiler):
            return AcaciaList(self.iterate())
        @cmethod_of(self, "values")
        @axe.chop
        def _values(compiler):
            return AcaciaList(self.values())
        @cmethod_of(self, "has")
        @axe.chop
        @axe.arg("key", axe.AnyValue())
        def _has(compiler, key: AcaciaExpr):
            return BoolLiteral(self._get_key(key) in self.data)
        @cmethod_of(self, "size")
        @axe.chop
        def _size(compiler):
            return IntLiteral(len(self.data))
        @cmethod_of(self, "get")
        @axe.chop
        @axe.arg("key", axe.AnyValue())
        @axe.arg("default", axe.AnyValue(), default=NoneLiteral())
        def _get(compiler, key: AcaciaExpr, default: AcaciaExpr):
            res = self.get(key)
            if res is None:
                res = default
            return res

    def iterate(self) -> List[ConstExpr]:
        return [k for k, _ in self.data.values()]

    def values(self) -> List[ConstExpr]:
        return [v for _, v in self.data.values()]

    def _get_key(self, key: AcaciaExpr):
        try:
            hash_ = key.hash()
        except InvalidOpError:
            raise Error(ErrorType.INVALID_MAP_KEY)
        else:
            return (type(key.data_type), hash_)

    def set(self, key: AcaciaExpr, value: ConstExpr):
        py_key = self._get_key(key)
        self.data[py_key] = (key, value)

    def get(self, key: AcaciaExpr) -> Optional[ConstExpr]:
        py_key = self._get_key(key)
        if py_key in self.data:
            _, v = self.data[py_key]
            return v
        return None

    def pop(self, key: AcaciaExpr) -> Optional[ConstExpr]:
        py_key = self._get_key(key)
        _, v = self.data.pop(py_key, (None, None))
        return v

    def items(self) -> Iterable[Tuple[ConstExpr, ConstExpr]]:
        return self.data.values()

    def to_ctexpr(self):
        return CTMap(list2ct(self.iterate()),
                     list2ct(self.values()))

class CTConstMap(CTObj):
    cdata_type = ctdt_constmap

    def __init__(self, keys: Iterable["CTExpr"],
                 values: Iterable["CTExpr"]):
        super().__init__()
        self.data: Dict[Hashable, Tuple[CTObj, CTObjPtr]] = {}
        for key, value in zip(keys, values):
            self.set(key, value)

        @cmethod_of(self, "__ct_getitem__")
        @axe.chop
        @axe.arg("key", axe.Constant())
        def _getitem(compiler, key: CTObj):
            res = self.get(key)
            if res is None:
                raise Error(ErrorType.MAP_KEY_NOT_FOUND)
            return abs(res)
        @cmethod_of(self, "copy")
        @axe.chop
        def _copy(compiler):
            return CTMap(self.citerate(), self.values())
        @cmethod_of(self, "keys")
        @axe.chop
        def _keys(compiler):
            return CTList(self.citerate())
        @cmethod_of(self, "values")
        @axe.chop
        def _values(compiler):
            return CTList(self.values())
        @cmethod_of(self, "get")
        @axe.chop
        @axe.arg("key", axe.Constant())
        @axe.arg("default", axe.Constant(), default=NoneLiteral())
        def _get(compiler, key: CTObj, default: CTObj):
            res = self.get(key)
            if res is None:
                res = default
            return abs(res)
        @cmethod_of(self, "has")
        @axe.chop
        @axe.arg("key", axe.Constant())
        def _has(compiler, key: CTObj):
            return BoolLiteral(self.get_key(key) in self.data)
        @cmethod_of(self, "size")
        @axe.chop
        def _size(compiler):
            return IntLiteral(len(self.data))

    def items(self) -> Iterable[Tuple[CTObj, CTObjPtr]]:
        return self.data.values()

    def citerate(self) -> List[CTObj]:
        return [k for k, _ in self.data.values()]

    def values(self) -> List["CTExpr"]:
        return [v for _, v in self.data.values()]

    @staticmethod
    def get_key(key: "CTExpr"):
        try:
            return abs(key).chash()
        except InvalidOpError:
            raise Error(ErrorType.INVALID_MAP_KEY)

    def set(self, key: "CTExpr", value: "CTExpr"):
        key = abs(key)
        value = abs(value)
        py_key = self.get_key(key)
        if py_key in self.data:
            _, ptr = self.data[py_key]
            ptr.set(value)
        else:
            self.data[py_key] = (key, CTObjPtr(value))

    def get(self, key: "CTExpr") -> Optional[CTObjPtr]:
        py_key = self.get_key(key)
        if py_key in self.data:
            _, ptr = self.data[py_key]
            return ptr
        return None

    def to_rt(self):
        return Map((k.to_rt() for k in self.citerate()),
                   (abs(v).to_rt() for v in self.values()))

class CTMap(CTConstMap):
    cdata_type = ctdt_map

    def __init__(self, keys: Iterable["CTExpr"],
                 values: Iterable["CTExpr"]):
        super().__init__(keys, values)
        @cmethod_of(self, "__ct_getitem__", runtime=False)
        @axe.chop
        @axe.arg("key", axe.Constant())
        def _getitem(compiler, key: "CTExpr"):
            res = self.get(key)
            if res is None:
                raise Error(ErrorType.MAP_KEY_NOT_FOUND)
            return res
        @cmethod_of(self, "update", runtime=False)
        @axe.chop
        @axe.arg("other", axe.MapOf(axe.Constant(), axe.Constant()))
        def _update(compiler, other: Dict[CTObj, CTObj]):
            for k, v in other.items():
                self.set(k, v)
        @cmethod_of(self, "pop", runtime=False)
        @axe.chop
        @axe.arg("key", axe.Constant())
        @axe.arg("default", axe.Constant(), default=None)
        def _pop(compiler, key: CTObj, default: Optional[CTObj]):
            res = self.pop(key)
            if res is None:
                if default is None:
                    raise Error(ErrorType.MAP_KEY_NOT_FOUND)
                else:
                    res = abs(default)
            return res
        @cmethod_of(self, "set_default", runtime=False)
        @axe.chop
        @axe.arg("key", axe.Constant())
        @axe.arg("default", axe.Constant(), default=NoneLiteral())
        def _set_default(compiler, key: CTObj, default: CTObj):
            res = self.get(key)
            if res is None:
                self.set(key, default)
                res = default
            return abs(res)
        @cmethod_of(self, "clear", runtime=False)
        @axe.chop
        def _clear(compiler):
            self.data.clear()

    def pop(self, key: "CTExpr") -> Optional[CTObj]:
        py_key = self.get_key(key)
        if py_key in self.data:
            _, ptr = self.data.pop(py_key)
            return abs(ptr)
        return None



"""Tokenizer (Lexer) for Acacia."""

__all__  = ['TokenType', 'Token', 'Tokenizer']

from typing import (
    Union, List, TextIO, Tuple, Dict, Optional, NamedTuple, Any, TYPE_CHECKING
)
import enum
import string

from acaciamc.error import *
from acaciamc.constants import COLORS, COLORS_NEW, FUNCTION_PATH_CHARS

if TYPE_CHECKING:
    from acaciamc.tools.versionlib import VERSION_T

UNICODE_ESCAPES = {'x': 2, 'u': 4, 'U': 8}
FONTS = {
    "reset": "r",
    "bold": "l",
    "italic": "o",
    "obfuscated": "k",
}

class TokenType(enum.Enum):
    """
    Token types.
    Value convension:
    1. If length of value is 1 or 2 and the first character is not an
    identifier character ([a-zA-Z0-9_]) or anything special (like '#'),
    then a token is registered automatically. (e.g. '=' or '->')
    2. If value is defined between `true` and `false`, then it is
    registered as a keyword.
    """
    # block
    indent = 'INDENT'
    dedent = 'DEDENT'
    new_line = 'NEW_LINE'
    lparen = '('
    rparen = ')'
    lbracket = '['
    rbracket = ']'
    lbrace = '{'
    rbrace = '}'
    colon = ':'
    comma = ','
    end_marker = 'END_MARKER'  # end of file
    # command and string
    command_begin = 'COMMAND_BEGIN'
    string_begin = 'STRING_BEGIN'
    text_body = 'TEXT_BODY'  # value: str
    dollar_lbrace = 'DOLLAR_LBRACE'
    command_end = 'COMMAND_END'
    string_end = 'STRING_END'
    # operator
    ## binary operators
    plus = '+'
    minus = '-'
    star = '*'
    slash = '/'
    mod = '%'
    ## compare operators
    equal_to = '=='
    unequal_to = '!='
    greater = '>'
    less = '<'
    greater_equal = '>='
    less_equal = '<='
    ## augmented assign
    plus_equal = '+='
    minus_equal = '-='
    times_equal = '*='
    divide_equal = '/='
    mod_equal = '%='
    ## other
    equal = '='
    arrow = '->'
    point = '.'
    at = '@'
    walrus = ':='
    ampersand = '&'
    # expressions
    ## literal
    integer = 'INTEGER'  # value: int
    float_ = 'FLOAT'  # value: float
    interface_path = 'INTERFACE_PATH'  # value: str
    ## id
    identifier = 'IDENTIFIER'  # value: str
    # keywords
    true = 'True'  # must at beginning of keyword part
    def_ = 'def'
    interface = 'interface'
    inline = 'inline'
    entity = 'entity'
    extends = 'extends'
    self = 'self'
    if_ = 'if'
    elif_ = 'elif'
    else_ = 'else'
    while_ = 'while'
    pass_ = 'pass'
    and_ = 'and'
    or_ = 'or'
    not_ = 'not'
    result = 'result'
    import_ = 'import'
    as_ = 'as'
    from_ = 'from'
    none = 'None'
    for_ = 'for'
    in_ = 'in'
    struct = 'struct'
    virtual = 'virtual'
    override = 'override'
    const = 'const'
    static = 'static'
    new = 'new'
    false = 'False'  # must at end of keyword part

KEYWORDS: Dict[str, TokenType] = {}
def _kw_builder():
    # NOTE this relies on that keywords in `TokenType` enum are all
    # between `true` and `false`.
    token_types = tuple(TokenType)
    for i in range(
        token_types.index(TokenType.true),
        token_types.index(TokenType.false) + 1
    ):
        token_type = token_types[i]
        KEYWORDS[token_type.value] = token_type
_kw_builder()
del _kw_builder

BRACKETS = {
    TokenType.lparen: TokenType.rparen,
    TokenType.lbracket: TokenType.rbracket,
    TokenType.lbrace: TokenType.rbrace
}
RB2LB = {v: k for k, v in BRACKETS.items()}

def is_idstart(c: str) -> bool:
    """Return if given character can start an identifier in Acacia."""
    return c.isalpha() or c == '_'

def is_idcontinue(c: str) -> bool:
    """Return if given character is valid in an Acacia identifier."""
    return is_idstart(c) or c in string.digits

class Token(NamedTuple):
    type: TokenType
    lineno: int
    col: int
    value: Any = None

    def __str__(self) -> str:
        v = self.type.value
        if not v.isupper():
            v = repr(v)
        if self.value is not None:
            v += ' (%s)' % self.value
        return v

    def __repr__(self):
        if self.value is None:
            value_str = ''
        else:
            value_str = '(%s)' % (self.value,)
        return '<Token %s%s at %s:%s>' % (
            self.type.name, value_str,
            self.lineno, self.col
        )

class _FormattedStrManager:
    def __init__(self, owner: 'Tokenizer',
                 begin_tok: TokenType, end_tok: TokenType) -> None:
        self.lineno = owner.current_lineno
        self.col = owner.current_col
        self.owner = owner
        self.begin_tok = begin_tok
        self.end_tok = end_tok
        self._texts: List[str] = []
        self._text_ln = 0
        self._text_col = 0
        self._tokens: List[Token] = [
            Token(self.begin_tok, self.lineno, self.col)
        ]

    def _dump_texts(self):
        if self._texts:
            self._tokens.append(Token(
                TokenType.text_body,
                self._text_ln, self._text_col, "".join(self._texts)
            ))
            self._texts.clear()

    def add_text(self, text: str):
        if not self._texts:
            self._text_ln = self.owner.current_lineno
            self._text_col = self.owner.current_col
        self._texts.append(text)

    def add_token(self, token: Token):
        self._dump_texts()
        self._tokens.append(token)

    def get_tokens(self) -> List[Token]:
        res = self._tokens.copy()
        self._tokens.clear()
        return res

    def finish(self):
        self._dump_texts()
        self._tokens.append(self.owner.gen_token(self.end_tok))

class _CommandManager(_FormattedStrManager):
    def __init__(self, owner: 'Tokenizer') -> None:
        super().__init__(owner, TokenType.command_begin, TokenType.command_end)

class _StrLiteralManager(_FormattedStrManager):
    def __init__(self, owner: 'Tokenizer') -> None:
        super().__init__(owner, TokenType.string_begin, TokenType.string_end)

class _BracketFrame(NamedTuple):
    type: TokenType
    pos: Tuple[int, int]
    cmd_fexpr: bool = False
    str_fexpr: bool = False

class Tokenizer:
    def __init__(self, src: TextIO, mc_version: "VERSION_T"):
        """src: source code"""
        self.src = src
        self.mc_version = mc_version
        self.current_char = ''
        self.current_lineno = 0
        self.current_col = 0
        self.position = 0  # string pointer
        self.buffer_tokens: List[Token] = []
        self.bracket_stack: List[_BracketFrame] = []
        self.indent_record: List[int] = [0]
        self.indent_len: int = 1  # always = len(self.indent_record)
        self.has_content = False  # if this logical line produced token
        self.last_line_continued = False
        self.inside_command: Optional[_CommandManager] = None
        self.in_command_fexpr = False
        self.string_stack: List[_StrLiteralManager] = []
        self.in_string_fexpr = False  # in fexpr of the most inner string
        self.continued_command = False
        self.continued_comment = False
        self.continued_comment_pos: Union[None, Tuple[int, int]] = None
        self.last_is_interface = False

    def error(self, err_type: ErrorType, lineno=None, col=None, **kwargs):
        if lineno is None:
            lineno = self.current_lineno
        if col is None:
            col = self.current_col
        err = Error(err_type, **kwargs)
        err.location.linecol = (lineno, col)
        raise err

    def forward(self):
        """Read next char and push pointer."""
        if self.position >= self.line_len:
            self.current_char = None
            self.current_col = self.line_len + 1
        else:
            self.current_char = self.current_line[self.position]
            self.position += 1
            self.current_col += 1

    def peek(self, offset=0):
        """Peek the following char without pushing pointer."""
        pos = self.position + offset
        if pos < self.line_len:
            return self.current_line[pos]
        return None

    def get_next_token(self):
        """Get the next token."""
        while not self.buffer_tokens:
            self.buffer_tokens = self.parse_line()
        return self.buffer_tokens.pop(0)

    def is_in_bracket(self) -> bool:
        return bool(self.bracket_stack)

    def parse_line(self) -> List[Token]:
        """Parse a line."""
        self.current_line = self.src.readline()
        self.line_len = len(self.current_line)
        self.current_lineno += 1
        self.current_col = 0
        self.position = 0
        res: List[Token] = []
        self.forward()
        # read indent
        if not (self.continued_command
                or self.continued_comment
                or self.last_line_continued
                or self.is_in_bracket()):
            # Generate indent token if this is not a continued line
            # (start of a new logical line).
            self.has_content = False
            prespaces = 0
            while self.current_char == ' ':
                prespaces += 1
                self.forward()
        else:
            prespaces = -1
            if self.continued_comment:
                self.handle_long_comment()
        while True:
            if self.inside_command and not self.in_command_fexpr:
                if self.continued_command:
                    res.extend(self.handle_long_command())
                    # If dollar lbrace is not closed, then the long
                    # command should not have ended. Instead the
                    # tokenizer would keep tokenizing as normal.
                    # e.g. `/* ${ */` results in unexpected token '*'.
                else:
                    res.extend(self.handle_command())
            if self.string_stack and not self.in_string_fexpr:
                res.extend(self.handle_string())
            # skip spaces
            self.skip_spaces()
            # check end of line
            if self.current_char in ("\n", "\\", None):
                break
            # comment
            if self.current_char == "#":
                if self.peek() == "*":
                    self.forward()
                    self.forward()
                    self.continued_comment_pos = (self.current_lineno,
                                                  self.current_col)
                    self.handle_long_comment()
                    continue
                else:
                    self.skip_comment()
                    break
            # start
            ## path after `interface` keyword
            if self.last_is_interface:
                if self.current_char == '"':
                    # Quoted path
                    self.enter_string()
                    self.forward()
                else:
                    # Unquoted path
                    res.append(self.handle_interface_path())
                self.last_is_interface = False
                continue
            ## special tokens
            ok = True
            if self.current_char in string.digits:
                res.append(self.handle_number())
            elif is_idstart(self.current_char):
                id_token = self.handle_name()
                res.append(id_token)
                # Special case: `interface` keyword, since after it goes
                # a path.
                if id_token.type is TokenType.interface:
                    self.last_is_interface = True
            elif self.current_char == '/' and not (self.has_content or res):
                # Only read when "/" is first token of this logical line.
                self.inside_command = _CommandManager(self)
                self.forward()  # skip "/"
                if self.current_char == '*':
                    self.continued_command = True
                    self.forward()  # skip "*"
            elif self.current_char == '"':
                self.forward()  # skip '"'
                self.enter_string()
            else:
                ok = False
            if ok:
                continue
            ## try two-char token
            peek = self.peek()
            if peek is not None:
                twochars = self.current_char + peek
                try:
                    token_type = TokenType(twochars)
                except ValueError:
                    pass
                else:
                    # It is indeed a two-char token!
                    res.append(self.gen_token(token_type))
                    self.forward()
                    self.forward()
                    continue
            ## try one-char token
            try:
                token_type = TokenType(self.current_char)
            except ValueError:
                # We've run out of possibilities.
                self.error(ErrorType.INVALID_CHAR, char=self.current_char)
            else:
                if token_type in BRACKETS:
                    self.bracket_stack.append(_BracketFrame(
                        token_type,
                        (self.current_lineno, self.current_col)
                    ))
                elif token_type in RB2LB:
                    if not self.bracket_stack:
                        self.error(ErrorType.UNMATCHED_BRACKET,
                                   char=self.current_char)
                    expect = RB2LB[token_type]
                    got_f = self.bracket_stack.pop()
                    got = got_f.type
                    if got is not expect:
                        self.error(ErrorType.UNMATCHED_BRACKET_PAIR,
                                   open=got.value, close=token_type.value)
                    if got is TokenType.lbrace:
                        if got_f.cmd_fexpr:
                            self.in_command_fexpr = False
                        elif got_f.str_fexpr:
                            self.in_string_fexpr = False
                res.append(self.gen_token(token_type))
                self.forward()
        # Now self.current_char is either '\n', '\\' or None (EOF)
        if (
            # ${} in single line command can't use implicit line continuation.
            (self.in_command_fexpr and not self.continued_command)
            # And so is ${} in string literal.
            or self.in_string_fexpr
        ):
            self.error(ErrorType.UNCLOSED_FEXPR)
        backslash = False
        eof = False
        if res:
            self.has_content = True
        if self.current_char == '\n':
            if self.has_content and not (
                self.continued_comment or self.continued_command
                or self.is_in_bracket()
            ):
                res.append(self.handle_logical_newline())
            self.last_line_continued = False
        elif self.current_char is None:
            if self.continued_comment:
                self.error(ErrorType.UNCLOSED_LONG_COMMENT,
                           *self.continued_comment_pos)
            if self.continued_command:
                if self.in_command_fexpr:
                    self.error(ErrorType.UNCLOSED_FEXPR)
                self.error(ErrorType.UNCLOSED_LONG_COMMAND,
                           self.inside_command.lineno,
                           self.inside_command.col)
            if self.is_in_bracket():
                br = self.bracket_stack[-1]
                self.error(ErrorType.UNCLOSED_BRACKET,
                           *br.pos, char=br.type.value)
            if self.has_content:
                res.append(self.handle_logical_newline())
            eof = True
        elif self.current_char == '\\':
            self.forward()  # skip "\\"
            if self.current_char is None:
                self.error(ErrorType.EOF_AFTER_CONTINUATION)
            if self.current_char != '\n':
                self.error(ErrorType.CHAR_AFTER_CONTINUATION)
            backslash = self.last_line_continued = True
        # Count indent: if this physical line is not a continued line,
        # and (this physical line has content or ends with backslash)
        if (res or backslash) and prespaces != -1:
            res[:0] = self.handle_indent(prespaces)
        # Add a fake line to clean up if we reach end of file
        if eof:
            self.current_lineno += 1
            self.current_col = 0
            res.extend(self.handle_indent(0))  # dump DEDENTs
            res.append(self.gen_token(TokenType.end_marker))
        return res

    def gen_token(self, *args, **kwargs):
        """Generate a token at current pos."""
        return Token(
            lineno=self.current_lineno, col=self.current_col,
            *args, **kwargs
        )

    def skip_spaces(self):
        """Skip white spaces."""
        while self.current_char == ' ':
            self.forward()

    def skip_comment(self):
        """Skip a single-line # comment."""
        while self.current_char is not None and self.current_char != '\n':
            self.forward()

    def enter_string(self):
        """Enter a string literal."""
        self.string_stack.append(_StrLiteralManager(self))
        self.in_string_fexpr = False

    def handle_long_comment(self):
        # we want to leave current_char on next char after "#",
        # so use previous_char
        previous_char = ''
        while not (previous_char == '*' and self.current_char == '#'):
            if self.current_char in ("\n", None):
                self.continued_comment = True
                return
            previous_char = self.current_char
            self.forward()
        self.forward()  # skip char "#"
        self.continued_comment = False

    def handle_logical_newline(self):
        """Generate a logical NEWLINE token and do checks."""
        if self.last_is_interface:
            self.error(ErrorType.INTERFACE_PATH_EXPECTED)
        return self.gen_token(TokenType.new_line)

    def handle_indent(self, spaces: int) -> List[Token]:
        """Generate INDENT and DEDENT."""
        ln = self.current_lineno
        tokens = []
        if spaces > self.indent_record[-1]:
            self.indent_record.append(spaces)
            self.indent_len += 1
            tokens.append(Token(TokenType.indent, lineno=ln, col=0))
        elif spaces in self.indent_record:
            i = self.indent_record.index(spaces)
            dedent_count = self.indent_len - 1 - i
            self.indent_record = self.indent_record[:i+1]
            self.indent_len -= dedent_count
            tokens.extend(Token(TokenType.dedent, lineno=ln, col=0)
                          for _ in range(dedent_count))
        else:
            self.error(ErrorType.INVALID_DEDENT)
        return tokens

    @staticmethod
    def _isdecimal(char: Union[str, None]):
        return char is not None and char in string.digits

    def handle_number(self):
        """Read an INTEGER or a FLOAT token."""
        res = []
        ln, col = self.current_lineno, self.current_col
        ## decide base
        base = 10
        peek = self.peek()
        if self.current_char == '0' and peek is not None:
            peek = peek.upper()
            if peek in "BOX":
                if peek == 'B':
                    base = 2
                elif peek == 'O':
                    base = 8
                elif peek == 'X':
                    base = 16
                # skip 2 chars ("0x" "0b" "0o")
                self.forward()
                self.forward()
        ## read
        valid_chars = '0123456789ABCDEF'[:base]
        while ((self.current_char is not None)
               and (self.current_char.upper() in valid_chars)):
            res.append(self.current_char)
            self.forward()
        ## convert string to number
        if (self.current_char == "."
            and base == 10
            and self._isdecimal(self.peek())
        ):  # float
            self.forward()
            res.append(".")
            while self._isdecimal(self.current_char):
                res.append(self.current_char)
                self.forward()
            value = float(''.join(res))
            return Token(TokenType.float_, value=value, lineno=ln, col=col)
        else:  # integer
            if not res:
                self.error(ErrorType.INTEGER_REQUIRED, base=base)
            value = int(''.join(res), base=base)
            return Token(TokenType.integer, value=value, lineno=ln, col=col)

    def handle_name(self):
        """Read a keyword or an IDENTIFIER token."""
        chars = []
        ln, col = self.current_lineno, self.current_col
        while is_idcontinue(self.current_char):
            chars.append(self.current_char)
            self.forward()
        name = ''.join(chars)
        token_type = KEYWORDS.get(name)
        if token_type is None:  # IDENTIFIER
            return Token(TokenType.identifier, value=name, lineno=ln, col=col)
        # Keyword
        return Token(token_type, lineno=ln, col=col)

    def _font2char(self, word: str) -> Optional[str]:
        if word in COLORS:
            return COLORS[word]
        if word in COLORS_NEW and self.mc_version >= (1, 19, 80):
            return COLORS_NEW[word]
        if word in FONTS:
            return FONTS[word]
        return None

    def _read_escapable_char(self) -> str:
        """Read current char as an escapable one (in string or command)
        and skip the read char(s). Return the handled char(s).
        """
        first = self.current_char
        self.forward()
        # not escapable
        if first != '\\':
            return first
        # escapable
        second = self.current_char
        if second == '\\':  # backslash itself
            self.forward()  # skip second backslash
            return second
        elif second == '#':  # font
            self.forward()  # skip '#'
            if self.current_char != '(':
                return '\xA7'
            start_ln, start_col = self.current_lineno, self.current_col
            res: List[str] = []
            cur_spec: List[str] = []
            self.forward()  # skip '('
            while True:
                if self.current_char is None or self.current_char == '\n':
                    self.error(ErrorType.UNCLOSED_FONT,
                               lineno=start_ln, col=start_col)
                if self.current_char in (',', ')'):
                    font = ''.join(cur_spec).strip()
                    cur_spec.clear()
                    ch = self._font2char(font)
                    if ch is None:
                        self.error(ErrorType.INVALID_FONT, font=font,
                                   lineno=start_ln, col=start_col)
                    res.append(f"\xA7{ch}")
                    if self.current_char == ')':
                        break
                else:
                    cur_spec.append(self.current_char)
                self.forward()
            self.forward()  # skip ')'
            return ''.join(res)
        ## NOTE '\n' should be passed directly to MC
        ## because MC use '\n' escape too
        elif second in UNICODE_ESCAPES:  # unicode number
            def _err():
                self.error(ErrorType.INVALID_UNICODE_ESCAPE,
                           escape_char=second)
            self.forward()  # skip '\\'
            code = []
            length = UNICODE_ESCAPES[second]
            for _ in range(length):
                if self.current_char not in string.hexdigits:
                    _err()
                code.append(self.current_char)
                self.forward()
            unicode = int(''.join(code), base=16)
            if unicode >= 0x110000:
                _err()
            return chr(unicode)
        # when escape can't be recognized, just return "\\"
        # and the `second` char will be handled later
        return first  # (here first == '\\')

    def handle_string(self) -> List[Token]:
        """Help read a string literal."""
        mgr = self.string_stack[-1]
        while self.current_char != '"':
            # check None and \n
            if (self.current_char is None) or (self.current_char == '\n'):
                self.error(ErrorType.UNCLOSED_QUOTE,
                           lineno=mgr.lineno, col=mgr.col)
            if self.current_char == '\\' and self.peek() == '"':
                # special escape in strings
                mgr.add_text('"')
                self.forward()
                self.forward()
                continue
            if self._fexpr_unit(mgr, is_cmd=False):
                break
        else:
            self.forward()  # skip last '"'
            mgr.finish()
            self.string_stack.pop()
            self.in_string_fexpr = bool(self.string_stack)
        return mgr.get_tokens()

    def _fexpr_unit(self, mgr: _FormattedStrManager, is_cmd: bool) -> bool:
        """
        Helper for parsing formatted expression in string and command.
        Read a character / escape sequence / start of a formatted
        expression. Return True if we got a formatted expression.
        is_cmd: True if we are in a command, False if we are in a string
        """
        peek = self.peek()
        do_break = False
        if self.current_char == '\\' and peek == '$':
            # special escape in commands
            mgr.add_text('$')
            self.forward()  # skip "\\"
            self.forward()  # skip "$"
        elif self.current_char == '$' and peek == '{':
            # formatted expression
            mgr.add_token(self.gen_token(TokenType.dollar_lbrace))
            self.forward()  # skip "$"
            self.bracket_stack.append(_BracketFrame(
                TokenType.lbrace,
                (self.current_lineno, self.current_col),
                **{("cmd_fexpr" if is_cmd else "str_fexpr"): True}
            ))  # Make sure position points to "{", not "$"
            self.forward()  # skip "{"
            if is_cmd:
                self.in_command_fexpr = True
            else:
                self.in_string_fexpr = True
            do_break = True
        else:  # a normal character
            mgr.add_text(self._read_escapable_char())
        return do_break

    def handle_long_command(self) -> List[Token]:
        """Help read a multi-line /*...*/ command. Return the tokens."""
        mgr = self.inside_command
        while not (self.current_char == '*' and self.peek() == '/'):
            # check whether we reach end of line
            if self.current_char in ("\n", None):
                # replace "\n" with " " (space)
                mgr.add_text(' ')
                break
            if self._fexpr_unit(mgr, is_cmd=True):
                break
        else:
            mgr.finish()
            self.forward()  # skip "*"
            self.forward()  # skip "/"
            self.continued_command = False
            self.inside_command = None
        return mgr.get_tokens()

    def handle_command(self) -> List[Token]:
        """Help read a single line command. Return the tokens."""
        mgr = self.inside_command
        while self.current_char is not None and self.current_char != '\n':
            if self._fexpr_unit(mgr, is_cmd=True):
                break
        else:
            mgr.finish()
            self.inside_command = None
        return mgr.get_tokens()

    def handle_interface_path(self) -> Token:
        """Help read an interface path."""
        ln, col = self.current_lineno, self.current_col
        # Read as long as current char is in FUNCTION_PATH_CHARS. Note
        # that Acacia removes support for parenthesis characters.
        chars = []
        while (
            self.current_char in FUNCTION_PATH_CHARS
            and self.current_char not in ('(', ')')
        ):
            chars.append(self.current_char)
            self.forward()
        if not chars:
            self.error(ErrorType.INTERFACE_PATH_EXPECTED)
        path = ''.join(chars)
        return Token(TokenType.interface_path, lineno=ln, col=col, value=path)



"""Builtin integer.

Three types of values are used to represent an Acacia int object:
- IntLiteral: a literal integer like `3` (for const folding)
- IntVar: describe a value stored in a Minecraft score
  which is usually an Acacia variable
- IntOpGroup: describe complicated expressions with operators

Q: WHY NOT just use one class to show everything???
A: With these classes apart, we can design a different operator logic for
   each kind of expression.
    e.g. 1 + 1 can be folded and directly get 2 when IntLiteral is apart.
    e.g. 1 + a can be optimized when IntVar is apart
   So the purpose is to optimize the output

This shows the priority of these classes
IntOpGroup > IntVar > IntLiteral
A class can only handle operation with other objects with lower priority
e.g. Literal can only handle operations with Literal
If a class can't handle an operation `self.xxx(other)`, other's method
`other.rxxx(self)` is used.
"""

__all__ = [
    # Type related
    'IntType', 'IntDataType',
    # Expression
    'IntLiteral', 'IntVar', 'IntOpGroup',
    # Utils
    'to_IntVar'
]

from typing import TYPE_CHECKING, List, Tuple, Callable, Optional, Dict, Union
from functools import partialmethod
import operator

from .types import Type
from .boolean import (
    CompareBase, BoolDataType, BoolLiteral, ScbMatchesCompare, to_BoolVar
)
from acaciamc.error import *
from acaciamc.constants import INT_MIN, INT_MAX
from acaciamc.tools import axe, cmethod_of
from acaciamc.ast import Operator, COMPOP_INVERT
from acaciamc.mccmdgen.expr import *
from acaciamc.mccmdgen.datatype import (
    DefaultDataType, Storable, SupportsEntityField
)
from acaciamc.mccmdgen.ctexpr import CTDataType
from acaciamc.mccmdgen.utils import InvalidOpError, unreachable
import acaciamc.mccmdgen.cmds as cmds

if TYPE_CHECKING:
    from acaciamc.compiler import Compiler
    from acaciamc.mccmdgen.cmds import _ExecuteSubcmd
    from acaciamc.mccmdgen.ctexpr import CTObj

STR2SCBOP = {
    "+": cmds.ScbOp.ADD_EQ,
    "-": cmds.ScbOp.SUB_EQ,
    "*": cmds.ScbOp.MUL_EQ,
    "/": cmds.ScbOp.DIV_EQ,
    "%": cmds.ScbOp.MOD_EQ,
    "<": cmds.ScbOp.MIN,
    ">": cmds.ScbOp.MAX,
    "=": cmds.ScbOp.ASSIGN
}
STR2METHOD = {
    '+': "add", '-': "sub",
    '*': "mul", '/': "div", '%': "mod"
}
COMPOP2SCBOP = {
    Operator.greater: cmds.ScbCompareOp.GT,
    Operator.greater_equal: cmds.ScbCompareOp.GTE,
    Operator.less: cmds.ScbCompareOp.LT,
    Operator.less_equal: cmds.ScbCompareOp.LTE,
    Operator.equal_to: cmds.ScbCompareOp.EQ
}
COMPOP2PYOP = {
    Operator.equal_to: operator.eq,
    Operator.unequal_to: operator.ne,
    Operator.greater: operator.gt,
    Operator.less: operator.lt,
    Operator.greater_equal: operator.ge,
    Operator.less_equal: operator.le
}

def remainder(a: int, b: int) -> int:
    """C-style % operator, which MC uses."""
    res = abs(a) % abs(b)
    return res if a >= 0 else -res

def c_int_div(a: int, b: int) -> int:
    """C-style / operator, which MC uses."""
    if (a >= 0) != (b >= 0) and a % b:
        return a // b + 1
    else:
        return a // b

STR2PYINTOP: Dict[str, Callable[[int, int], int]] = {
    '+': int.__add__, '-': int.__sub__, '*': int.__mul__,
    '/': c_int_div, '%': remainder
}

def _int_rawtextify(self: AcaciaExpr, compiler: "Compiler"):
    """A generic `rawtextify` implementation for int."""
    dependencies, var = to_IntVar(self, compiler)
    return [cmds.RawtextScore(var.slot)], dependencies

class IntDataType(DefaultDataType, Storable, SupportsEntityField):
    name = "int"

    def new_var(self, compiler) -> "IntVar":
        return IntVar.new(compiler)

    def new_entity_field(self, compiler):
        return {"scoreboard": compiler.add_scoreboard()}

    def new_var_as_field(self, entity, scoreboard: str) -> "IntVar":
        return IntVar(cmds.ScbSlot(entity.to_str(), scoreboard))

ctdt_int = CTDataType("int")

class IntType(Type):
    def do_init(self):
        self.attribute_table.set('MAX', IntLiteral(INT_MAX))
        self.attribute_table.set('MIN', IntLiteral(INT_MIN))
        @cmethod_of(self, "__new__")
        @axe.chop
        @axe.arg("b", BoolDataType)
        @axe.slash
        def _new(compiler, b):
            """
            int(b: bool, /) -> int
            Returns 1 if `b` is True, or 0 otherwise.
            """
            if isinstance(b, BoolLiteral):
                return IntLiteral(int(b.value))
            # Fallback: convert `b` to `BoolVar`,
            # Since 0 is used to store False, 1 is for True, just
            # "cast" it to `IntVar`.
            dependencies, bool_var = to_BoolVar(b, compiler)
            res = IntVar(bool_var.slot)
            res.is_temporary = True
            return res, dependencies

    def datatype_hook(self):
        return IntDataType()

    def cdatatype_hook(self):
        return ctdt_int

class IntCompare(CompareBase):
    """
    A boolean that stores comparison between 2 integer expressions.
    This only stores one comparison, unlike the AST node `CompareOp`,
    so "a > b > 1" is stored in two `IntCompare`s.
    """
    def __init__(
        self, left: AcaciaExpr, operator: Operator,
        right: AcaciaExpr
    ):
        super().__init__()
        self.left = left
        self.operator = operator
        self.right = right

    def as_execute(self, compiler) -> Tuple[CMDLIST_T, List["_ExecuteSubcmd"]]:
        res_dependencies = []  # return[0]
        res_main = []  # return[1]
        dependency, var_left = to_IntVar(self.left, compiler)
        res_dependencies.extend(dependency)
        dependency, var_right = to_IntVar(self.right, compiler)
        res_dependencies.extend(dependency)
        # != is specially handled, because Minecraft does not provide
        # such syntax like 'if score ... != ...' while other 5
        # operators are built-in in MC.
        if self.operator is Operator.unequal_to:
            res_main.append(cmds.ExecuteScoreComp(
                var_left.slot, var_right.slot,
                cmds.ScbCompareOp.EQ, invert=True
            ))
        else:  # for other 5 operators
            res_main.append(cmds.ExecuteScoreComp(
                var_left.slot, var_right.slot, COMPOP2SCBOP[self.operator]
            ))
        return res_dependencies, res_main

    # Unary operator
    def unarynot(self, compiler):
        return IntCompare(self.left, COMPOP_INVERT[self.operator], self.right)

class IntLiteral(ConstExprCombined):
    """Represents a literal integer.
    The purpose of these class is to implement constant folding
    which calculate the value of constant expressions
    in compile time (e.g. compiler can convert "2 + 3" to "5").
    """
    cdata_type = ctdt_int

    def __init__(self, value: int):
        super().__init__(IntDataType())
        self.value = value
        # check overflow
        if not INT_MIN <= value <= INT_MAX:
            raise Error(ErrorType.INT_OVERFLOW)

    def cstringify(self) -> str:
        return str(self.value)

    def chash(self):
        return self.value

    def export(self, var: "IntVar", compiler):
        return [cmds.ScbSetConst(var.slot, self.value)]

    def ccompare(self, op, other: Union[AcaciaExpr, "CTObj"]) -> bool:
        if isinstance(other, IntLiteral):
            return COMPOP2PYOP[op](self.value, other.value)
        raise InvalidOpError

    def implicitcast(self, type_, compiler):
        from .float_ import FloatDataType, Float
        if type_.matches_cls(FloatDataType):
            return Float(float(self.value))
        raise InvalidOpError

    ## UNARY OPERATORS

    def cunarypos(self):
        return self

    def cunaryneg(self):
        return IntLiteral(-self.value)

    ## BINARY OPERATORS

    def _bin_op(self, op: str, other: Union[AcaciaExpr, "CTObj"]):
        if isinstance(other, IntLiteral):
            try:
                v = STR2PYINTOP[op](self.value, other.value)
            except ArithmeticError as err:
                raise Error(ErrorType.CONST_ARITHMETIC, message=str(err))
            return IntLiteral(v)
        raise InvalidOpError

    cadd = partialmethod(_bin_op, '+')
    csub = partialmethod(_bin_op, '-')
    cmul = partialmethod(_bin_op, '*')
    cdiv = partialmethod(_bin_op, '/')
    cmod = partialmethod(_bin_op, '%')

class IntVar(VarValue):
    """An integer variable."""
    def __init__(self, slot: cmds.ScbSlot):
        super().__init__(IntDataType())
        self.slot = slot

    @classmethod
    def new(cls, compiler: "Compiler", tmp=False):
        alloc = compiler.allocate_tmp if tmp else compiler.allocate
        return cls(alloc())

    def export(self, var: "IntVar", compiler):
        return [cmds.ScbOperation(cmds.ScbOp.ASSIGN, var.slot, self.slot)]

    def compare(self, op, other, compiler):
        if not other.data_type.matches_cls(IntDataType):
            raise InvalidOpError
        if isinstance(other, IntLiteral):
            return ScbMatchesCompare([], self.slot, op, other.value)
        return IntCompare(self, op, other)

    def swap(self, other: "IntVar", compiler):
        return [cmds.ScbOperation(cmds.ScbOp.SWAP, self.slot, other.slot)]

    rawtextify = _int_rawtextify

    ## UNARY OPERATORS

    def unarypos(self, compiler):
        return self

    def unaryneg(self, compiler):
        return IntOpGroup.from_intexpr(self).unaryneg(compiler)

    ## BINARY (SELF ... OTHER) OPERATORS

    def _bin_op(self, name: str, other, compiler):
        # `name` is method name
        if isinstance(other, (IntLiteral, IntVar)):
            meth = getattr(IntOpGroup.from_intexpr(self), name)
            return meth(other, compiler)
        raise InvalidOpError

    add = partialmethod(_bin_op, 'add')
    sub = partialmethod(_bin_op, 'sub')
    mul = partialmethod(_bin_op, 'mul')
    div = partialmethod(_bin_op, 'div')
    mod = partialmethod(_bin_op, 'mod')

    ## BINARY (OTHER ... SELF) OPERATORS
    # only `IntLiteral` might call rxxx of self
    # so just convert self to `IntOpGroup` and let this handle operation

    def _r_bin_op(self, name, other, compiler):
        if isinstance(other, IntLiteral):
            meth = getattr(IntOpGroup.from_intexpr(other), name)
            return meth(self, compiler)
        raise InvalidOpError

    radd = partialmethod(_r_bin_op, 'add')
    rsub = partialmethod(_r_bin_op, 'sub')
    rmul = partialmethod(_r_bin_op, 'mul')
    rdiv = partialmethod(_r_bin_op, 'div')
    rmod = partialmethod(_r_bin_op, 'mod')

    # AUGMENTED ASSIGN `ixxx`

    def _iadd_sub(self, operator: str, other,
                  compiler: "Compiler") -> CMDLIST_T:
        """Implementation of iadd and isub."""
        if isinstance(other, IntLiteral):
            cls = cmds.ScbAddConst if operator == '+' else cmds.ScbRemoveConst
            return [cls(self.slot, other.value)]
        elif isinstance(other, IntVar):
            return [cmds.ScbOperation(
                STR2SCBOP[operator], self.slot, other.slot
            )]
        elif isinstance(other, IntOpGroup):
            return self._aug_IntOpGroup(other, operator, compiler)
        raise InvalidOpError

    def _imul_div_mod(self, operator: str, other,
                      compiler: "Compiler") -> CMDLIST_T:
        """Implementation of imul, idiv, imod."""
        if isinstance(other, IntLiteral):
            const = compiler.add_int_const(other.value)
            return [cmds.ScbOperation(
                STR2SCBOP[operator], self.slot, const.slot
            )]
        elif isinstance(other, IntVar):
            return [cmds.ScbOperation(
                STR2SCBOP[operator], self.slot, other.slot
            )]
        elif isinstance(other, IntOpGroup):
            return self._aug_IntOpGroup(other, operator, compiler)
        raise InvalidOpError

    def _aug_IntOpGroup(self, other: "IntOpGroup", operator: str, compiler):
        """Implementation for augmented assigns where `other` is
        `IntOpGroup`. `operator` is "+", "-", etc.
        """
        # In this condition, we convert a (op)= b to a = a (op) b
        ## Calculate
        value = getattr(self, STR2METHOD[operator])(other, compiler)
        ## Export
        tmp = IntVar.new(compiler, tmp=True)
        res = value.export(tmp, compiler)
        res.extend(tmp.export(self))
        return res

    iadd = partialmethod(_iadd_sub, '+')
    isub = partialmethod(_iadd_sub, '-')
    imul = partialmethod(_imul_div_mod, '*')
    idiv = partialmethod(_imul_div_mod, '/')
    imod = partialmethod(_imul_div_mod, '%')

class IntOp:
    # This is intended to be immutable
    def scb_did_read(self, slot: cmds.ScbSlot) -> bool:
        # Override-able
        return False

    def scb_did_assign(self, slot: cmds.ScbSlot) -> bool:
        # Override-able
        return False

    def resolve(self, var: IntVar) -> CMDLIST_T:
        """
        Override this or `resolve_full`.
        The returned list won't be modified.
        """
        raise NotImplementedError(
            "Must override either `resolve` or `resolve_full`"
        )

    def resolve_full(self, var: IntVar, compiler: "Compiler") -> CMDLIST_T:
        """
        Override this or `resolve`.
        The returned list won't be modified.
        """
        return self.resolve(var)

class IntSetConst(IntOp):
    def __init__(self, value: int) -> None:
        self.value = value

    def resolve(self, var: IntVar) -> CMDLIST_T:
        return [cmds.ScbSetConst(var.slot, self.value)]

class IntRandom(IntOp):
    def __init__(self, min_: int, max_: int) -> None:
        self.min = min_
        self.max = max_

    def resolve(self, var: IntVar) -> CMDLIST_T:
        return [cmds.ScbRandom(var.slot, self.min, self.max)]

class IntOpConst(IntOp):
    def __init__(self, op: str, value: int) -> None:
        self.op = op
        self.value = value

    def resolve_full(self, var: IntVar, compiler: "Compiler") -> CMDLIST_T:
        if self.op == "+" or self.op == "-":
            cls = cmds.ScbAddConst if self.op == "+" else cmds.ScbRemoveConst
            return [cls(var.slot, self.value)]
        c = compiler.add_int_const(self.value)
        return [cmds.ScbOperation(STR2SCBOP[self.op], var.slot, c.slot)]

class IntOpVar(IntOp):
    def __init__(self, op: str, slot: cmds.ScbSlot) -> None:
        self.op = STR2SCBOP[op]
        self.slot = slot

    def scb_did_read(self, slot: cmds.ScbSlot) -> bool:
        return slot == self.slot

    def resolve(self, var: IntVar) -> CMDLIST_T:
        return [cmds.ScbOperation(self.op, var.slot, self.slot)]

class IntSetVar(IntOpVar):
    def __init__(self, slot: cmds.ScbSlot) -> None:
        super().__init__("=", slot)

class IntOpSelf(IntOp):
    def __init__(self, op: str) -> None:
        self.op = op

    def resolve(self, var: IntVar) -> CMDLIST_T:
        return [cmds.ScbOperation(STR2SCBOP[self.op], var.slot, var.slot)]

class IntCmdOp(IntOp):
    def __init__(self, commands: CMDLIST_T) -> None:
        self.commands = commands

    def scb_did_read(self, slot: cmds.ScbSlot) -> bool:
        for c in self.commands:
            if c.scb_did_read(slot):
                return True
        return False

    def scb_did_assign(self, slot: cmds.ScbSlot) -> bool:
        for c in self.commands:
            if c.scb_did_assign(slot):
                return True
        return False

    def resolve(self, var: IntVar) -> CMDLIST_T:
        return self.commands

class IntOpGroup(AcaciaExpr):
    """An `IntOpGroup` stores complex integer operations."""
    def __init__(self, init: Optional[IntOp]):
        super().__init__(IntDataType())
        self.ops: List[IntOp] = []
        if init is not None:
            self.ops.append(init)

    @classmethod
    def from_intexpr(cls, init: AcaciaExpr) -> "IntOpGroup":
        """
        Create an `IntOpGroup` that sets the value to `init`.
        `init` must be int type.
        """
        if isinstance(init, IntOpGroup):
            return init
        res = cls(None)
        if isinstance(init, IntLiteral):
            res.add_op(IntSetConst(init.value))
        elif isinstance(init, IntVar):
            res.add_op(IntSetVar(init.slot))
        else:
            unreachable()
        return res

    def export(self, var: IntVar, compiler):
        need_tmp = False
        for op in self.ops:
            if op.scb_did_read(var.slot) or op.scb_did_assign(var.slot):
                need_tmp = True
                break
        if need_tmp:
            tmp = IntVar.new(compiler, tmp=True)
        else:
            tmp = var
        res = []
        for op in self.ops:
            res.extend(op.resolve_full(tmp, compiler))
        if need_tmp:
            res.extend(tmp.export(var, compiler))
        return res

    def copy(self):
        res = IntOpGroup(init=None)
        res.ops = self.ops.copy()
        return res

    def compare(self, op, other, compiler):
        if not other.data_type.matches_cls(IntDataType):
            raise InvalidOpError
        if isinstance(other, IntLiteral):
            commands, var = to_IntVar(self, compiler)
            return ScbMatchesCompare(commands, var.slot, op, other.value)
        return IntCompare(self, op, other)

    rawtextify = _int_rawtextify

    def add_op(self, op: IntOp):
        self.ops.append(op)

    ## UNARY OPERATORS

    def unarypos(self, compiler):
        return self

    def unaryneg(self, compiler):
        # -expr = expr * (-1)
        return self.mul(compiler.add_int_const(-1))

    ## BINARY (SELF ... OTHER) OPERATORS

    def _bin_op(self, operator: str, other, compiler):
        """Implements binary operators (see `STR2SCBOP`)."""
        res = self.copy()
        if isinstance(other, IntLiteral):
            res.add_op(IntOpConst(operator, other.value))
        elif isinstance(other, IntVar):
            res.add_op(IntOpVar(operator, other.slot))
        elif isinstance(other, IntOpGroup):
            tmp = IntVar.new(compiler, tmp=True)
            res.add_op(IntCmdOp(other.export(tmp, compiler)))
            res.add_op(IntOpVar(operator, tmp.slot))
        else:
            raise InvalidOpError
        return res

    add = partialmethod(_bin_op, '+')
    sub = partialmethod(_bin_op, '-')
    mul = partialmethod(_bin_op, '*')
    div = partialmethod(_bin_op, '/')
    mod = partialmethod(_bin_op, '%')

    ## BINARY (OTHER ... SELF) OPERATORS
    # `other` in self.rxxx may be IntLiteral or IntVar

    def _r_add_mul(self, name: str, other, compiler):
        # a (+ or *) b is b (+ or *) a
        if isinstance(other, (IntLiteral, IntVar)):
            return getattr(self, name)(other)
        raise InvalidOpError

    def _r_sub_div_mod(self, other, name, compiler):
        # Convert `other` to `IntOpGroup` and use this to handle this
        # operation.
        if isinstance(other, (IntLiteral, IntVar)):
            meth = getattr(IntOpGroup.from_intexpr(other), name)
            return meth(self, compiler)
        raise InvalidOpError

    radd = partialmethod(_r_add_mul, 'add')
    rmul = partialmethod(_r_add_mul, 'mul')
    rsub = partialmethod(_r_sub_div_mod, 'sub')
    rdiv = partialmethod(_r_sub_div_mod, 'div')
    rmod = partialmethod(_r_sub_div_mod, 'mod')

# Utils
def to_IntVar(expr: AcaciaExpr, compiler, tmp=True) \
        -> Tuple[CMDLIST_T, IntVar]:
    """Convert any integer expression to an `IntVar` and some commands.
    return[0]: the commands to run
    return[1]: the `IntVar`
    """
    if isinstance(expr, IntVar):
        return [], expr
    else:
        var = IntVar.new(compiler, tmp=tmp)
        return expr.export(var, compiler), var



"""Error definition for Acacia."""

__all__ = ['SourceLocation', 'ErrorType', 'Error', 'ErrFrame', 'traced_call']

from typing import Optional, Tuple, NamedTuple, List, Callable, Union

from acaciamc.localization import LocalizedEnum, localize

class SourceLocation:
    """Represent a location in a source file for showing errors."""
    def __init__(
            self, file: Optional[str] = None,
            linecol: Optional[Tuple[int, int]] = None):
        self.file = file
        self.linecol = linecol

    def file_set(self) -> bool:
        return self.file is not None

    def linecol_set(self) -> bool:
        return self.linecol is not None

    def __str__(self) -> str:
        if self.file is None:
            return "<unknown>"
        if self.linecol is None:
            return self.file
        return "%s:%d:%d" % (self.file, self.linecol[0], self.linecol[1])

class ErrorType(LocalizedEnum):
    # Tokenizer
    INVALID_CHAR = "error.errortype.invalidchar"
    INTEGER_REQUIRED = "error.errortype.integerrequired"
    INT_OVERFLOW = "error.errortype.intoverflow"
    UNCLOSED_LONG_COMMENT = "error.errortype.unclosedlongcomment"
    UNCLOSED_LONG_COMMAND = "error.errortype.unclosedlongcommand"
    UNCLOSED_QUOTE = "error.errortype.unclosedquote"
    UNCLOSED_FEXPR = "error.errortype.unclosedfexpr"
    UNCLOSED_FONT = "error.errortype.unclosedfont"
    UNCLOSED_BRACKET = "error.errortype.unclosedbracket"
    UNMATCHED_BRACKET = "error.errortype.unmatchedbracket"
    UNMATCHED_BRACKET_PAIR = "error.errortype.unmatchedbracketpair"
    INVALID_UNICODE_ESCAPE = "error.errortype.invalidunicodeescape"
    INVALID_FONT = "error.errortype.invalidfont"
    CHAR_AFTER_CONTINUATION = "error.errortype.charaftercontinuation"
    EOF_AFTER_CONTINUATION = "error.errortype.eofaftercontinuation"
    INVALID_DEDENT = "error.errortype.invaliddedent"
    INTERFACE_PATH_EXPECTED = "error.errortype.interfacepathexpected"
    # Parser
    UNEXPECTED_TOKEN = "error.errortype.unexpectedtoken"
    EMPTY_BLOCK = "error.errortype.emptyblock"
    DONT_KNOW_ARG_TYPE = "error.errortype.dontknowargtype"
    DUPLICATE_ARG_DEF = "error.errortype.duplicateargdef"
    POSITIONED_ARG_AFTER_KEYWORD = "error.errortype.positionedargafterkeyword"
    NONDEFAULT_ARG_AFTER_DEFAULT = "error.errortype.nondefaultargafterdefault"
    INVALID_FUNC_PORT = "error.errortype.invalidfuncport"
    INVALID_VARDEF_STMT = "error.errortype.invalidvardefstmt"
    NONSTATIC_CONST_METHOD = "error.errortype.nonstaticconstmethod"
    # Command Generator
    NAME_NOT_DEFINED = "error.errortype.namenotdefined"
    HAS_NO_ATTRIBUTE = "error.errortype.hasnoattribute"
    MODULE_NO_ATTRIBUTE = "error.errortype.modulenoattribute"
    INVALID_OPERAND = "error.errortype.invalidoperand"
    INVALID_BOOLOP_OPERAND = "error.errortype.invalidboolopoperand"
    UNSUPPORTED_VAR_TYPE = "error.errortype.unsupportedvartype"
    UNSUPPORTED_ARG_TYPE = "error.errortype.unsupportedargtype"
    UNSUPPORTED_RESULT_TYPE = "error.errortype.unsupportedresulttype"
    UNSUPPORTED_EFIELD_TYPE = "error.errortype.unsupportedefieldtype"
    UNSUPPORTED_SFIELD_TYPE = "error.errortype.unsupportedsfieldtype"
    UNSUPPORTED_EFIELD_IN_STRUCT = "error.errortype.unsupportedefieldinstruct"
    SHADOWED_NAME = "error.errortype.shadowedname"
    WRONG_ASSIGN_TYPE = "error.errortype.wrongassigntype"
    WRONG_REF_TYPE = "error.errortype.wrongreftype"
    WRONG_CONST_TYPE = "error.errortype.wrongconsttype"
    WRONG_ARG_TYPE = "error.errortype.wrongargtype"
    WRONG_RESULT_TYPE = "error.errortype.wrongresulttype"
    WRONG_IF_CONDITION = "error.errortype.wrongifcondition"
    WRONG_WHILE_CONDITION = "error.errortype.wrongwhilecondition"
    CANT_REF = "error.errortype.cantref"
    CANT_REF_ARG = "error.errortype.cantrefarg"
    CANT_REF_RESULT = "error.errortype.cantrefresult"
    NOT_CONST = "error.errortype.notconst"
    NOT_CONST_NAME = "error.errortype.notconstname"
    NOT_CONST_ATTR = "error.errortype.notconstattr"
    ARG_NOT_CONST = "error.errortype.argnotconst"
    NONREF_ARG_DEFAULT_NOT_CONST = "error.errortype.nonrefargdefaultnotconst"
    ARG_DEFAULT_NOT_CONST = "error.errortype.argdefaultnotconst"
    RESULT_NOT_CONST = "error.errortype.resultnotconst"
    ELEMENT_NOT_CONST = "error.errortype.elementnotconst"
    MULTIPLE_RESULTS = "error.errortype.multipleresults"
    NON_RT_RESULT = "error.errortype.nonrtresult"
    NON_RT_NAME = "error.errortype.nonrtname"
    NON_RT_ATTR = "error.errortype.nonrtattr"
    ENDLESS_WHILE_LOOP = "error.errortype.endlesswhileloop"
    INVALID_TYPE_SPEC = "error.errortype.invalidtypespec"
    INVALID_ETEMPLATE = "error.errortype.invalidetemplate"
    INVALID_STEMPLATE = "error.errortype.invalidstemplate"
    UNMATCHED_ARG_DEFAULT_TYPE = "error.errortype.unmatchedargdefaulttype"
    # ARG_MULTIPLE_VALUES is used by both Generator and Parser
    ARG_MULTIPLE_VALUES = "error.errortype.argmultiplevalues"
    DUPLICATE_EFIELD = "error.errortype.duplicateefield"
    EFIELD_MULTIPLE_DEFS = "error.errortype.efieldmultipledefs"
    EMETHOD_MULTIPLE_DEFS = "error.errortype.emethodmultipledefs"
    MULTIPLE_VIRTUAL_METHOD = "error.errortype.multiplevirtualmethod"
    METHOD_ATTR_CONFLICT = "error.errortype.methodattrconflict"
    MULTIPLE_NEW_METHODS = "error.errortype.multiplenewmethods"
    CONST_NEW_METHOD = "error.errortype.constnewmethod"
    SFIELD_MULTIPLE_DEFS = "error.errortype.sfieldmultipledefs"
    MISSING_ARG = "error.errortype.missingarg"
    RESULT_OUT_OF_SCOPE = "error.errortype.resultoutofscope"
    SELF_OUT_OF_SCOPE = "error.errortype.selfoutofscope"
    NEW_OUT_OF_SCOPE = "error.errortype.newoutofscope"
    TOO_MANY_ARGS = "error.errortype.toomanyargs"
    UNEXPECTED_KEYWORD_ARG = "error.errortype.unexpectedkeywordarg"
    UNCALLABLE = "error.errortype.uncallable"
    NOT_ITERABLE = "error.errortype.notiterable"
    NO_GETITEM = "error.errortype.nogetitem"
    INVALID_ASSIGN_TARGET = "error.errortype.invalidassigntarget"
    INVALID_FEXPR = "error.errortype.invalidfexpr"
    INVALID_BIN_FUNC_ARG = "error.errortype.invalidbinfuncarg"
    CANT_CREATE_INSTANCE = "error.errortype.cantcreateinstance"
    CANT_CREATE_ENTITY = "error.errortype.cantcreateentity"
    ENTITY_NEW_RETURN_TYPE = "error.errortype.entitynewreturntype"
    CONST_ARITHMETIC = "error.errortype.constarithmetic"
    MRO = "error.errortype.mro"
    OVERRIDE_RESULT_MISMATCH = "error.errortype.overrideresultmismatch"
    OVERRIDE_RESULT_UNSTORABLE = "error.errortype.overrideresultunstorable"
    OVERRIDE_QUALIFIER = "error.errortype.overridequalifier"
    NOT_OVERRIDING = "error.errortype.notoverriding"
    INST_OVERRIDE_STATIC = "error.errortype.instoverridestatic"
    STATIC_OVERRIDE_INST = "error.errortype.staticoverrideinst"
    VIRTUAL_OVERRIDE_SIMPLE = "error.errortype.virtualoverridesimple"
    UNINITIALIZED_CONST = "error.errortype.uninitializedconst"
    INVALID_CONST_STMT = "error.errortype.invalidconststmt"
    POS_OFFSET_CTOR_ARG = "error.errortype.posoffsetctorarg"
    INVALID_POS_ALIGN = "error.errortype.invalidposalign"
    LIST_INDEX_OUT_OF_BOUNDS = "error.errortype.listindexoutofbounds"
    MAP_KEY_NOT_FOUND = "error.errortype.mapkeynotfound"
    INVALID_MAP_KEY = "error.errortype.invalidmapkey"
    LIST_MULTIMES_NON_LITERAL = "error.errortype.listmultimesnonliteral"
    INVALID_UPCAST = "error.errortype.invalidupcast"
    NEVER_RESULT = "error.errortype.neverresult"
    RESERVED_INTERFACE_PATH = "error.errortype.reservedinterfacepath"
    DUPLICATE_INTERFACE = "error.errortype.duplicateinterface"
    INTERFACE_PATH_EMPTY = "error.errortype.interfacepathempty"
    INTERFACE_PATH_SLASH_START = "error.errortype.interfacepathslashstart"
    INTERFACE_PATH_SLASH_END = "error.errortype.interfacepathslashend"
    INTERFACE_PATH_DOUBLE_SLASH = "error.errortype.interfacepathdoubleslash"
    INTERFACE_PATH_INVALID_CHAR = "error.errortype.interfacepathinvalidchar"
    # Compiler
    IO = "error.errortype.io"
    MODULE_NOT_FOUND = "error.errortype.modulenotfound"
    CIRCULAR_PARSE = "error.errortype.circularparse"
    # Any; should only be used by binary modules
    ANY = "error.errortype.any"

class ErrFrame(NamedTuple):
    location: SourceLocation
    msg: str
    note: Optional[str]

class Error(Exception):
    def __init__(self, err_type: ErrorType, **kwargs):
        self.error_args = kwargs
        self.type = err_type
        self.location = SourceLocation()
        self.frames: List[ErrFrame] = []
        super().__init__()

    def __str__(self):
        res = self.type.localized  # unformatted error
        res = res.format(**self.error_args)  # formatted
        res = '%s: %s' % (self.location, res)
        return res

    def full_msg(self) -> str:
        return (
            localize("error.fullmsg.compilererror")
            + ("\n" if self.frames else " ")
            + "".join(
                "  %s: %s" % (frame.location, frame.msg)
                + ("\n    %s" % frame.note if frame.note else "")
                + "\n"
                for frame in reversed(self.frames)
            )
            + str(self)
        )

    def add_frame(self, frame: ErrFrame):
        self.frames.append(frame)

def traced_call(
    func: Callable, location: Union[SourceLocation, str, None],
    source: Optional[SourceLocation], func_repr: str,
    *args, **kwds
):
    try:
        return func(*args, **kwds)
    except Error as err:
        if location is None:
            location = SourceLocation()
        elif isinstance(location, str):
            location = SourceLocation(file=location)
        if source is not None and source.file_set():
            note = localize("error.tracedcall.note") % source
        else:
            note = None
        err.add_frame(ErrFrame(
            location,
            localize("error.tracedcall.calling") % func_repr,
            note
        ))
        raise



"""Named tuple style compile-time immutable objects in Acacia."""

__all__ = ["named_tuple"]

from typing import Dict, NamedTuple, Tuple, Any, Type as PyType, Sequence
from types import new_class, SimpleNamespace

from acaciamc.mccmdgen.expr import *
from acaciamc.mccmdgen.datatype import DefaultDataType
from acaciamc.mccmdgen.ctexpr import CTDataType
from acaciamc.mccmdgen.utils import apply_decorators
from acaciamc.tools import axe, cmethod_of
from acaciamc.objects.types import Type

class WithOrigin(axe.ConverterContainer):
    def __init__(self, converter: axe._CONVERTER_T):
        super().__init__(converter)
        self.converter = converter

    def convert(self, origin):
        return origin, self.converter.convert(origin)

    def cconvert(self, origin):
        return origin, self.converter.cconvert(origin)

    def get_show_name(self) -> str:
        return self.converter.get_show_name()

class _NamedTupleInfo(NamedTuple):
    name: str
    cls_datatype: PyType[DefaultDataType]
    cls_expr: PyType[ConstExprCombined]
    cls_type: PyType[Type]
    cdata_type: CTDataType

_MISSING = object()

def named_tuple(
    name: str, fields: Sequence[Tuple[str, axe._CONVERTER_T]],
    defaults: Dict[str, Any]
) -> _NamedTupleInfo:
    field_names = [name for name, _ in fields]
    if any(s.startswith("__") for s in field_names):
        raise ValueError("Field names cannot start with '__'")
    def dt_body(ns):
        ns["name"] = name
    dt = new_class(f"NamedTupleDataType_{name}", (DefaultDataType,),
                   exec_body=dt_body)
    ctdt = CTDataType(name)
    def expr_body(ns):
        def __init__(self: AcaciaExpr, data: Dict[str, Tuple[Any, Any]]):
            super(expr, self).__init__(dt())
            self.fields = SimpleNamespace()
            for fname, (raw, converted) in data.items():
                setattr(self.fields, fname, converted)
                self.attribute_table.set(fname, raw)
        ns["__init__"] = __init__
        ns["cdata_type"] = ctdt
    expr = new_class(f"NamedTuple_{name}", (ConstExprCombined,),
                     exec_body=expr_body)
    ctor_args1 = [
        axe.arg(fname, WithOrigin(conv))
        for fname, conv in fields
        if fname not in defaults
    ]
    ctor_args2 = [
        axe.arg(
            fname, WithOrigin(conv),
            default=(
                defaults[fname],  # raw
                axe._get_convert(defaults[fname], conv)()  # converted
            )
        )
        for fname, conv in fields
        if fname in defaults
    ]
    evolve_args = [
        axe.arg(fname, WithOrigin(conv), default=_MISSING)
        for fname, conv in fields
    ]
    def type_body(ns):
        def do_init(self):
            @cmethod_of(self, "__new__")
            @axe.chop
            @apply_decorators(ctor_args1)
            @apply_decorators(ctor_args2)
            def _new(compiler, **kwds):
                return expr(kwds)
            @cmethod_of(self, "evolve")
            @axe.chop
            @axe.arg("__original", dt)
            @axe.slash
            @apply_decorators(evolve_args)
            def _evolve(compiler, __original: AcaciaExpr, **kwds):
                for fname in field_names:
                    if kwds[fname] is _MISSING:
                        kwds[fname] = (
                            __original.attribute_table.lookup(fname),  # raw
                            getattr(__original.fields, fname)  # converted
                        )
                return expr(kwds)
        def datatype_hook(self):
            return dt()
        def cdatatype_hook(self):
            return ctdt
        ns["do_init"] = do_init
        ns["datatype_hook"] = datatype_hook
        ns["cdatatype_hook"] = cdatatype_hook
    typecls = new_class(f"NamedTupleType_{name}", (Type,),
                        exec_body=type_body)
    return _NamedTupleInfo(name, dt, expr, typecls, ctdt)



"""Struct template."""

__all__ = ["StructTemplateDataType", "StructTemplate"]

from typing import List, Dict, Optional, TYPE_CHECKING

from .struct import StructDataType, Struct
from .functions import ConstructorFunction, BinaryFunction
from acaciamc.error import *
from acaciamc.tools import axe, resultlib
from acaciamc.mccmdgen.expr import *
from acaciamc.mccmdgen.datatype import DefaultDataType
from acaciamc.mccmdgen.ctexpr import CTDataType
from acaciamc.mccmdgen.utils import apply_decorators

if TYPE_CHECKING:
    from acaciamc.mccmdgen.datatype import Storable

class StructTemplateDataType(DefaultDataType):
    name = "struct_template"

ctdt_structtemplate = CTDataType("struct_template")

class StructTemplate(ConstExprCombined, ConstructorFunction):
    cdata_type = ctdt_structtemplate

    def __init__(self, name: str, field: Dict[str, "Storable"],
                 bases: List["StructTemplate"], source=None):
        super().__init__(StructTemplateDataType())
        self.name = name
        self.bases = bases
        self.field_types = field
        self.func_repr = name
        self.source = source
        # Merge attributes from ancestors.
        for base in bases:
            for name, type_ in base.field_types.items():
                if name in self.field_types:
                    raise Error(ErrorType.SFIELD_MULTIPLE_DEFS, attr=name)
                self.field_types[name] = type_

    def datatype_hook(self):
        return StructDataType(template=self)

    def is_subtemplate_of(self, template: "StructTemplate") -> bool:
        """Return whether `template` is sub-template of this.
        This itself is treated as its own sub-template.
        """
        if template is self:
            return True
        if template in self.bases:
            return True
        return any(base.is_subtemplate_of(template) for base in self.bases)

    def initialize(
        self, instance: "Struct", compiler,
        args: "ARGS_T", keywords: "KEYWORDS_T"
    ):
        @axe.chop
        @axe.star
        @apply_decorators([
            axe.arg(name, type_, default=None)
            for name, type_ in self.field_types.items()
        ])
        def _call_me(compiler, **fields: Optional[AcaciaExpr]):
            commands = []
            for name, value in fields.items():
                if value is None:
                    continue
                commands.extend(value.export(instance.vars[name], compiler))
            return resultlib.commands(commands)
        _, c = BinaryFunction(_call_me).call(args, keywords, compiler)
        return c



"""Command line interface of Acacia."""

__all__ = ["build_argparser", "get_config", "run", "main"]

import argparse
import os
import shutil
import sys

from acaciamc.error import Error as CompileError
from acaciamc.compiler import Compiler, Config
from acaciamc.localization import localize
from acaciamc.tokenizer import is_idstart, is_idcontinue

_NOTGIVEN = object()

def fatal(message: str):
    print(localize("cli.fatal") % message, file=sys.stderr)
    sys.exit(1)

def build_argparser():
    argparser = argparse.ArgumentParser(
        prog='acacia', description=localize("cli.description"),
    )
    argparser.add_argument(
        'file',
        help=localize("cli.argshelp.file")
    )
    argparser.add_argument(
        '-o', '--out', metavar='PATH',
        help=localize("cli.argshelp.out")
    )
    argparser.add_argument(
        "-v", "--mc-version", metavar="VERSION",
        help=localize("cli.argshelp.mcversion")
    )
    argparser.add_argument(
        '-e', '--education-edition',
        action='store_true',
        help=localize("cli.argshelp.educationedition")
    )
    argparser.add_argument(
        '-s', '--scoreboard', metavar='OBJECTIVE',
        help=localize("cli.argshelp.scoreboard")
    )
    argparser.add_argument(
        '-f', '--function-folder', metavar='PATH',
        help=localize("cli.argshelp.functionfolder")
    )
    argparser.add_argument(
        '-m', '--main-file', metavar='NAME',
        help=localize("cli.argshelp.mainfile")
    )
    argparser.add_argument(
        '-t', '--entity-tag', metavar="TAG",
        help=localize("cli.argshelp.entitytag")
    )
    argparser.add_argument(
        '-d', '--debug-comments',
        action='store_true',
        help=localize("cli.argshelp.debugcomments")
    )
    argparser.add_argument(
        "-O", "--no-optimize",
        action='store_true',
        help=localize("cli.argshelp.nooptimize")
    )
    argparser.add_argument(
        '-u', '--override-old',
        action='store_true',
        help=localize("cli.argshelp.overrideold")
    )
    argparser.add_argument(
        '-i', '--init-file', nargs='?', metavar='NAME', const=_NOTGIVEN,
        help=localize("cli.argshelp.initfile")
    )
    argparser.add_argument(
        '--internal-folder', metavar="NAME",
        help=localize("cli.argshelp.internalfolder")
    )
    argparser.add_argument(
        '--encoding', metavar="CODEC", default="utf-8",
        help=localize("cli.argshelp.encoding")
    )
    argparser.add_argument(
        '--verbose',
        action='store_true',
        help=localize("cli.argshelp.verbose")
    )
    argparser.add_argument(
        '--max-inline-file-size', metavar="SIZE", type=int,
        help=localize("cli.argshelp.maxinline")
    )
    return argparser

def check_id(name: str):
    """Raise ValueError if `name` is not a valid Acacia identifier."""
    if not name:
        raise ValueError(localize("cli.checkid.empty"))
    if not is_idstart(name[0]):
        raise ValueError(localize("cli.checkid.idstart") % name[0])
    for c in name:
        if not is_idcontinue(c):
            raise ValueError(localize("cli.checkid.idcontinue") % c)

def assert_id(name: str, option: str):
    """Make sure `name` is a valid Acacia identifier."""
    try:
        check_id(name)
    except ValueError as e:
        fatal(localize("cli.assertid.fatal")
              .format(option=option, msg=e.args[0]))

def get_config(args) -> Config:
    """Create a `Config` object from `args`."""
    kwds = {
        'debug_comments': bool(args.debug_comments),
        'education_edition': bool(args.education_edition),
        'optimizer': not bool(args.no_optimize),
        'encoding': args.encoding
    }
    if args.scoreboard:
        assert_id(args.scoreboard, '--scoreboard')
        kwds["scoreboard"] = args.scoreboard
    if args.function_folder:
        path = [p for p in args.function_folder.split("/") if p]
        for p in path:
            try:
                check_id(p)
            except ValueError as e:
                fatal(localize("cli.getconfig.invalidfunctionfolder")
                      .format(msg=e.args[0], name=p))
        kwds["root_folder"] = '/'.join(path)
    if args.main_file:
        assert_id(args.main_file, '--main-file')
        kwds["main_file"] = args.main_file
    if args.entity_tag:
        kwds["entity_tag"] = args.entity_tag
    if args.mc_version:
        numlist = args.mc_version.split(".")
        try:
            t = tuple(map(int, numlist))
            if len(t) <= 1:
                raise ValueError
            if any(v < 0 for v in t):
                raise ValueError
        except ValueError:
            fatal(localize("cli.getconfig.invalidmcversion") % args.mc_version)
        if t < (1, 19, 50):
            fatal(localize("cli.getconfig.mcversiontooold") % args.mc_version)
        kwds["mc_version"] = t
    if args.max_inline_file_size is not None:
        if args.max_inline_file_size < 0:
            fatal(localize("cli.getconfig.maxinlinetoolow")
                  % args.max_inline_file_size)
        kwds["max_inline_file_size"] = args.max_inline_file_size
    if args.init_file:
        kwds["split_init"] = True
        if args.init_file is not _NOTGIVEN:
            assert_id(args.init_file, '--init-file')
            kwds["init_file"] = args.init_file
    if args.internal_folder:
        assert_id(args.internal_folder, '--internal-folder')
        kwds["internal_folder"] = args.internal_folder
    return Config(**kwds)

def try_rmtree(path: str):
    """
    Delete `path` if this path exists.
    Report any error as `fatal`.
    """
    if os.path.exists(path):
        try:
            shutil.rmtree(path)
        except OSError as e:
            fatal(localize("cli.tryrmtree.failure")
                  .format(path=path, message=e.strerror))

def run(args):
    if not os.path.exists(args.file):
        fatal('file not found: %s' % args.file)
    if not os.path.isfile(args.file):
        fatal('not a file: %s' % args.file)

    if args.out:
        out_path = os.path.realpath(args.out)
    else:  # default out path: ./<name of source>.acaout
        out_path, _ = os.path.splitext(args.file)
        out_path = os.path.realpath(out_path) + '.acaout'

    cfg = get_config(args)

    if not os.path.exists(out_path):
        out_up = os.path.dirname(out_path)
        if not os.path.exists(out_up):
            fatal(localize("cli.run.outputnotfound") % out_up)
        os.mkdir(out_path)

    try:
        compiler = Compiler(args.file, cfg)
        if args.override_old:
            # Remove old output directory if -u is set and compilation
            # succeeded.
            try_rmtree(os.path.join(out_path, cfg.root_folder))
        compiler.output(out_path)
    except CompileError as err:
        fatal(err.full_msg())
    except Exception as err:
        import traceback
        if args.verbose:
            traceback.print_exc()
            print(file=sys.stderr)
            fatal(localize("cli.run.aboveunexpectederror"))
        else:
            fatal(
                localize("cli.run.unexpectederror")
                % traceback.format_exception_only(err)[-1].strip()
            )

def main():
    argparser = build_argparser()
    args = argparser.parse_args()
    run(args)



"""Entity of Acacia."""

__all__ = ['EntityDataType', 'TaggedEntity', 'EntityReference']

from typing import TYPE_CHECKING, List, Optional

from acaciamc.error import *
from acaciamc.mccmdgen.mcselector import MCSelector
from acaciamc.mccmdgen.datatype import Storable
from acaciamc.mccmdgen.expr import *
from acaciamc.mccmdgen.utils import InvalidOpError

if TYPE_CHECKING:
    from acaciamc.compiler import Compiler
    from acaciamc.mccmdgen.datatype import DataType
    from .entity_template import EntityTemplate

class EntityDataType(Storable):
    def __init__(self, template: "EntityTemplate"):
        super().__init__()
        self.template = template

    def __str__(self) -> str:
        return self.template.name

    @classmethod
    def name_no_generic(cls) -> str:
        return "entity"

    def matches(self, other: "DataType") -> bool:
        return (isinstance(other, EntityDataType) and
                other.template.is_subtemplate_of(self.template))

    def new_var(self, compiler) -> "TaggedEntity":
        return TaggedEntity.new_tag(self.template, compiler)

class _EntityBase(AcaciaExpr):
    def __init__(self, template: "EntityTemplate",
                 cast_to: Optional["EntityTemplate"] = None):
        super().__init__(EntityDataType(template))
        self.cast_template = cast_to
        self.template = template
        self.template.register_entity(self)

    def __str__(self) -> str:
        return self.to_str()

    def to_str(self) -> str:
        return self.get_selector().to_str()

    def stringify(self) -> str:
        return self.to_str()

    def rawtextify(self, compiler):
        # To override default behavior of calling `stringify`.
        raise InvalidOpError

    def get_selector(self) -> MCSelector:
        res = self._get_selector()
        if res.var == "e" or res.var == "a":
            res.limit(1)
        return res

    def _get_selector(self) -> MCSelector:
        # Caller owns the selector
        raise NotImplementedError

    def cast_to(self, template: "EntityTemplate"):
        raise NotImplementedError

    def export(self, var: "TaggedEntity", compiler):
        """
        NOTE this default implementation assumes that clearing the
        target tag does not affect this object.
        """
        cmds = var.clear()
        cmds.append("tag %s add %s" % (self, var.tag))
        return cmds

class EntityReference(_EntityBase):
    def __init__(self, selector: MCSelector, template: "EntityTemplate",
                 cast_to: Optional["EntityTemplate"] = None):
        self.selector = selector
        super().__init__(template, cast_to)

    @classmethod
    def from_other(cls, other: _EntityBase):
        return cls(other.get_selector(), other.template)

    def _get_selector(self) -> MCSelector:
        return self.selector.copy()

    def cast_to(self, template):
        return EntityReference(self.selector, self.template, template)

    def export(self, var: "TaggedEntity", compiler):
        # Check if clearing the target tag would affect this object
        if var.tag in self.selector.get_tags():
            # By definition, an entity object must represent exactly one
            # entity, which means that this entity must satisfy the
            # selector, or else it's an undefined bahaviour.
            # XXX add runtime check to see if the entity actually
            # satisfies the selector.
            return []
        # Default implementation
        return super().export(var, compiler)

class TaggedEntity(_EntityBase, VarValue):
    def __init__(self, tag: str, template: "EntityTemplate",
                 cast_to: Optional["EntityTemplate"] = None):
        self.tag = tag
        super().__init__(template, cast_to)

    def cast_to(self, template):
        return TaggedEntity(self.tag, self.template, template)

    def export(self, var: "TaggedEntity", compiler):
        # Check self-assignment
        if var.tag == self.tag:
            return []
        return super().export(var, compiler)

    def swap(self, other: "TaggedEntity", compiler):
        # Check self-swapping
        if other.tag == self.tag:
            return []
        return super().swap(other, compiler)

    @classmethod
    def new_tag(cls, template: "EntityTemplate", compiler: "Compiler",
                cast_to: Optional["EntityTemplate"] = None):
        """Create a new entity tag that points to no entity."""
        return cls(
            compiler.allocate_entity_tag(),
            template, cast_to
        )

    def _get_selector(self) -> MCSelector:
        res = MCSelector("e")
        res.tag(self.tag)
        return res

    def clear(self) -> List[str]:
        # Clear the reference to entity that the tag is pointing to.
        return ["tag %s remove %s" % (self, self.tag)]



"""Builtin None value."""

__all__ = ['NoneDataType', 'NoneLiteral']

from acaciamc.mccmdgen.expr import *
from acaciamc.mccmdgen.datatype import (
    DefaultDataType, Storable, SupportsEntityField
)
from acaciamc.mccmdgen.ctexpr import CTDataType

class NoneDataType(DefaultDataType, Storable, SupportsEntityField):
    name = 'None'

    def new_var(self, compiler) -> "NoneLiteral":
        return NoneLiteral()

    def new_entity_field(self, compiler) -> dict:
        return {}

    def new_var_as_field(self, entity) -> VarValue:
        return NoneLiteral()

ctdt_none = CTDataType("None")

class NoneLiteral(ConstExprCombined, VarValue):
    """Represents a literal None."""
    cdata_type = ctdt_none

    def __init__(self):
        super().__init__(NoneDataType())

    def export(self, var: "NoneLiteral", compiler):
        return []

    def swap(self, other: "NoneLiteral", compiler):
        return []

    def datatype_hook(self):
        """None as a type specifier represents the None type."""
        return NoneDataType()

    def cdatatype_hook(self) -> CTDataType:
        return ctdt_none

    def stringify(self) -> str:
        return 'None'



"""A convenience object that represents an absolute position.
Most important: it is compatible with both `Pos` and `Offset`.
"""

import acaciamc.mccmdgen.cmds as cmds
from acaciamc.mccmdgen.expr import *
from acaciamc.mccmdgen.ctexpr import CTDataType
from acaciamc.tools import axe, cmethod_of, transform_immutable
from acaciamc.constants import XYZ
from .types import Type
from .position import PosDataType, Position, ctdt_position
from .position_offset import (
    PosOffsetDataType, PosOffset, CoordinateType, ctdt_posoffset
)
from .float_ import Float

class AbsPosDataType(PosDataType, PosOffsetDataType):
    name = "AbsPos"

ctdt_abspos = CTDataType("AbsPos", (ctdt_posoffset, ctdt_position))

class AbsPosType(Type):
    def do_init(self):
        @cmethod_of(self, "__new__")
        @axe.chop
        @axe.arg("x", axe.PosXZ())
        @axe.arg("y", axe.LiteralFloat())
        @axe.arg("z", axe.PosXZ())
        def _new(compiler, x: float, y: float, z: float):
            return AbsPos(x, y, z)

    def datatype_hook(self):
        return AbsPosDataType()

    def cdatatype_hook(self):
        return ctdt_abspos

class AbsPos(Position, PosOffset):
    cdata_type = ctdt_abspos

    # XXX As you can see the inheritance is a bit weird, we can't even
    # call __init__ properly.
    def __init__(self, x: float, y: float, z: float):
        AcaciaExpr.__init__(self, AbsPosDataType())
        self._context = cmds.ExecuteEnv("positioned", "")
        self.context = [self._context]
        self.values = [x, y, z]
        self.value_types = [CoordinateType.ABSOLUTE for _ in range(3)]
        self.already_set = set()
        self._update()
        # for attr in ("dim", "local", "apply", "align"):
        #     self.attribute_table.delete(attr)

        @cmethod_of(self, "abs")
        @axe.chop
        @axe.arg("x", axe.Nullable(axe.PosXZ()), default=None)
        @axe.arg("y", axe.Nullable(axe.LiteralFloat()), default=None)
        @axe.arg("z", axe.Nullable(axe.PosXZ()), default=None)
        @transform_immutable(self)
        def _abs(self: "AbsPos", compiler, x, y, z):
            for i, value in enumerate((x, y, z)):
                if value is not None:
                    self.set_abs(i, value)
            return self
        @cmethod_of(self, "offset")
        @axe.chop
        @axe.arg("x", axe.Nullable(axe.LiteralFloat()), default=None)
        @axe.arg("y", axe.Nullable(axe.LiteralFloat()), default=None)
        @axe.arg("z", axe.Nullable(axe.LiteralFloat()), default=None)
        @transform_immutable(self)
        def _offset(self: "AbsPos", compiler, x, y, z):
            for i, value in enumerate((x, y, z)):
                if value is not None:
                    self.set_offset(i, value)
            return self

    def copy(self):
        return AbsPos(*self.values)

    def _update(self):
        self._context.args = str(self)
        for name, value in zip(XYZ, self.values):
            self.attribute_table.set(name, Float(value))

    def set_abs(self, i: int, value: float):
        self.values[i] = value
        self._update()

    def set_offset(self, i: int, value: float):
        self.values[i] += value
        self._update()



"""Base stuffs for generating codes of Acacia expressions."""

__all__ = [
    # Base classes
    'AcaciaExpr', 'VarValue', 'AcaciaCallable',
    'ConstExpr', 'ConstExprCombined',
    # Type checking
    'ARGS_T', 'KEYWORDS_T', 'CALLRET_T', 'ITERLIST_T', 'CMDLIST_T',
    # Expression-related routines
    'swap_exprs'
]

from typing import (
    List, Union, Dict, Tuple, Hashable, Optional, Sequence, TYPE_CHECKING
)
from abc import ABCMeta, abstractmethod

from acaciamc.error import *
from acaciamc.mccmdgen.symbol import SymbolTable
from acaciamc.mccmdgen.ctexpr import CTObj
from acaciamc.mccmdgen.utils import InvalidOpError
from acaciamc.mccmdgen import cmds
from acaciamc.localization import localize

if TYPE_CHECKING:
    from acaciamc.ast import Operator
    from acaciamc.compiler import Compiler
    from acaciamc.mccmdgen.ctexpr import CTExpr
    from acaciamc.mccmdgen.datatype import DataType

ARGS_T = List["AcaciaExpr"]  # Positional arguments
KEYWORDS_T = Dict[str, "AcaciaExpr"]  # Keyword arguments
CMDLIST_T = List[Union[str, cmds.Command]]
CALLRET_T = Tuple["AcaciaExpr", CMDLIST_T]  # Result
ITERLIST_T = List["AcaciaExpr"]

class AcaciaExpr:
    """Base class for EVERYTHING that represents an Acacia expression.
    NOTE Contributer Guide:
    There are 2 types of Acacia data types:
    - The "storable" types. These types of expressions are stored in
      Minecraft, and their values can be changed when running the code
      (since it is stored in MC).
      e.g. int, bool
    - The "unstorable" types. These types of expression cannot be stored
      in Minecraft, and is only tracked by the compiler. Therefore,
      their values can't be changed.
      e.g. str, module, function
    To define a new type, you need to:
     - Define a subclass of `DataType` that represents this type.
       To implement a simple type, just subclass `DefaultDataType`, in
       which you just need to specify `name` attribute.
     - Define at least one subclass of `AcaciaExpr`, which represents
       the objects of this type.
       To make your object callable in Acacia, subclass `AcaciaCallable`
       and implement `call` method.
       `stringify` is a special method that returns the string
       representation of this expression used in raw command
       substitution and string interpolation.
    Extra things for "storable" types to implement:
     - at least one (usually 1) class that is a subclass of `VarValue`,
       to represent this kind of value that is stored in Minecraft.
     - `export` method, called when user assigns the value of this
       expression to a `VarValue` of same type.
     - the class that defines this type (subclass of `DataType`) should
       subclass `Storable`. It will have to implement `new_var` method,
       to allow creating a new `VarValue` of this type.
    Take builtin "int" as an example:
     - `IntVar`, which holds a Minecraft score, is implemented.
     - All the AcaciaExprs of `int` type implements `export`.
     - `IntDataType` does have a `new_var` method, which creates a new
       `IntVar`.

    To implement operator for your type, here are some methods:
     - add, sub, mul, div, mod: represents binary +, -, *, /, %,
       respectively. There are also radd, rsub, rmul, rdiv, rmod,
       which are similar to those in Python.
     - unarypos, unaryneg, unarynot: represents unary +, - and "not",
       respectively.
     - iadd, isub, imul, idiv, imod: represents +=, -=, *=, /=, %=
       respectively.
    When you are not satisfied with input operand type, please raise
    `InvalidOpError`.
    """
    def __init__(self, type_: "DataType"):
        super().__init__()
        self.data_type = type_
        self.attribute_table = SymbolTable()

    def is_assignable(self) -> bool:
        """Return whether this expression is a lvalue at runtime."""
        return False

    def export(self, var: "VarValue", compiler: "Compiler") -> CMDLIST_T:
        """Return the commands that assigns value of `self` to `var`.
        Since we need a `VarValue` here, only "storable" types need
        to implement this.
        """
        raise NotImplementedError

    def stringify(self) -> str:
        """Return a string representation of this expression, used in
        raw commands and string interpolations. If not implemented, then
        the object can not be used for those purposes.
        """
        raise InvalidOpError

    def rawtextify(self, compiler: "Compiler") \
            -> Tuple[Sequence[cmds.RawtextComponent], CMDLIST_T]:
        """Used when this expression is formatted in a JSON rawtext. It
        should return a list of rawtext components and commands or raise
        `InvalidOpError`. If the error was raised, then this expression
        cannot be used in rawtext. The returned lists are not modified.
        The default implementation tries to return the string
        representation by calling `stringify`.
        """
        return [cmds.RawtextText(self.stringify())], []

    def iterate(self) -> ITERLIST_T:
        """Implements for-in iteration. Should return an iterable
        of `AcaciaExpr`, in which values are bound to "for" variable
        one by one. If not implemented, then the object can't be
        iterated in a for-in structure.
        NOTE that for-in on an entity group is handled completely
        differently by the generator, so it should never implement this.
        """
        raise InvalidOpError

    def datatype_hook(self) -> "DataType":
        """When this expression is used as a type specifier, this
        method is called to obtain the `DataType`.
        """
        raise InvalidOpError

    def hash(self) -> Hashable:
        """When this expression is used as a map's key, this method
        is called to obtain the hash value.
        """
        raise InvalidOpError

    def compare(self, op: "Operator", other: "AcaciaExpr",
                compiler: "Compiler") -> "AcaciaExpr":
        """
        Implement comparison operators for this expression.
        Return value should be a boolean expression.
        """
        raise InvalidOpError

    def implicitcast(self, type_: "DataType", compiler: "Compiler") \
            -> "AcaciaExpr":
        """
        Perform implicit type conversion to `type_`. The retuned
        expression need to be of `type_` type.
        NOTE This conversion happens only when this expression is used
        as an argument in a call to a function that is written in
        Acacia.
        """
        raise InvalidOpError

    def add(self, other: "AcaciaExpr", compiler: "Compiler") -> "AcaciaExpr":
        raise InvalidOpError
    def sub(self, other: "AcaciaExpr", compiler: "Compiler") -> "AcaciaExpr":
        raise InvalidOpError
    def mul(self, other: "AcaciaExpr", compiler: "Compiler") -> "AcaciaExpr":
        raise InvalidOpError
    def div(self, other: "AcaciaExpr", compiler: "Compiler") -> "AcaciaExpr":
        raise InvalidOpError
    def mod(self, other: "AcaciaExpr", compiler: "Compiler") -> "AcaciaExpr":
        raise InvalidOpError

    def radd(self, other: "AcaciaExpr", compiler: "Compiler") -> "AcaciaExpr":
        raise InvalidOpError
    def rsub(self, other: "AcaciaExpr", compiler: "Compiler") -> "AcaciaExpr":
        raise InvalidOpError
    def rmul(self, other: "AcaciaExpr", compiler: "Compiler") -> "AcaciaExpr":
        raise InvalidOpError
    def rdiv(self, other: "AcaciaExpr", compiler: "Compiler") -> "AcaciaExpr":
        raise InvalidOpError
    def rmod(self, other: "AcaciaExpr", compiler: "Compiler") -> "AcaciaExpr":
        raise InvalidOpError

    def unarypos(self, compiler: "Compiler") -> "AcaciaExpr":
        raise InvalidOpError
    def unaryneg(self, compiler: "Compiler") -> "AcaciaExpr":
        raise InvalidOpError
    def unarynot(self, compiler: "Compiler") -> "AcaciaExpr":
        raise InvalidOpError

    def iadd(self, compiler: "Compiler") -> CMDLIST_T:
        raise InvalidOpError
    def isub(self, compiler: "Compiler") -> CMDLIST_T:
        raise InvalidOpError
    def imul(self, compiler: "Compiler") -> CMDLIST_T:
        raise InvalidOpError
    def idiv(self, compiler: "Compiler") -> CMDLIST_T:
        raise InvalidOpError
    def imod(self, compiler: "Compiler") -> CMDLIST_T:
        raise InvalidOpError

class ConstExpr(AcaciaExpr, metaclass=ABCMeta):
    @abstractmethod
    def to_ctexpr(self) -> "CTExpr":
        pass

class VarValue(AcaciaExpr):
    """
    `VarValue`s are special `AcaciaExpr`s that can be assigned to. It
    represents a "variable" of a storable type and is sometimes used as
    a temporary (with `is_temporary` set to `True`, in which case it
    cannot be assigned).
    e.g. 1 + 2 -> IntLiteral(3) -> Unassignable
    e.g. a -> IntVar(ScbSlot("acacia3", "acacia")) -> Assignable
    e.g. scb("x", "scb") -> IntVar(ScbSlot("x", "scb")) -> Assignable
    e.g. bool -> Type -> Unassignable
    """
    is_temporary = False  # used as a temporary and is read-only

    def swap(self, other: "VarValue", compiler: "Compiler") -> CMDLIST_T:
        raise InvalidOpError

    def is_assignable(self) -> bool:
        return not self.is_temporary

class AcaciaCallable(AcaciaExpr, metaclass=ABCMeta):
    """Acacia expressions that are callable."""
    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)
        self.source: Optional[SourceLocation] = None
        self.func_repr = localize("objects.function.unknown")

    def call_withframe(
            self, args: ARGS_T, keywords: KEYWORDS_T, compiler: "Compiler",
            location: Optional[Union[SourceLocation, str]] = None
        ) -> CALLRET_T:
        """
        Call this expression, and add this to error frame if an error
        occurs.
        """
        return traced_call(
            self.call, location, self.source, self.func_repr,
            args, keywords, compiler
        )

    @abstractmethod
    def call(self, args: ARGS_T, keywords: KEYWORDS_T,
             compiler: "Compiler") -> CALLRET_T:
        """
        Call this expression.
        Return value:
         1st element: Result of this call
         2nd element: Commands to run
        """
        pass

class ConstExprCombined(ConstExpr, CTObj):
    def __init_subclass__(cls) -> None:
        def ct2rt(func):
            def newfunc(self, *args, **kwds):
                if "compiler" in kwds:
                    kwds.pop("compiler")
                else:
                    args = list(args)
                    args.pop()
                return func(self, *args, **kwds)
            return newfunc
        for meth in (
            'add', 'sub', 'mul', 'div', 'mod',
            'radd', 'rsub', 'rmul', 'rdiv', 'rmod',
            'unarypos', 'unaryneg', 'unarynot'
        ):
            rtfunc = getattr(cls, meth)
            defaultrt = getattr(AcaciaExpr, meth)
            if rtfunc is not defaultrt:
                continue  # already defined
            cmeth = f"c{meth}"
            ctfunc = getattr(cls, cmeth)
            defaultct = getattr(CTObj, cmeth)
            if ctfunc is not defaultct:
                setattr(cls, meth, ct2rt(ctfunc))

    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)
        self.attributes = self.attribute_table

    def to_ctexpr(self):
        return self

    def to_rt(self):
        return self

    def stringify(self):
        return self.cstringify()

    def compare(self, op: "Operator", other: AcaciaExpr, compiler):
        from acaciamc.objects.boolean import BoolLiteral
        b = self.ccompare(op, other)
        assert isinstance(b, bool)
        return BoolLiteral(b)

    def hash(self):
        return self.chash()

def swap_exprs(x: VarValue, y: VarValue, compiler: "Compiler") -> CMDLIST_T:
    """Swap two `VarValue`s that are assignable to each other."""
    try:
        res = x.swap(y, compiler)
    except InvalidOpError:
        # Fall back
        tmp = x.data_type.new_var(compiler)
        res = [
            *x.export(tmp, compiler),
            *y.export(x, compiler),
            *tmp.export(y, compiler)
        ]
    return res



"""Tools about return value of binary functions."""

__all__ = ["Result", "commands", "literal", "nothing"]

from typing import NamedTuple, Union, TYPE_CHECKING

# `import acaciamc.objects as objects` won't work in 3.6
# because of a Python bug (see https://bugs.python.org/issue23203)
from acaciamc import objects
from acaciamc.mccmdgen.utils import unreachable

if TYPE_CHECKING:
    from acaciamc.mccmdgen.expr import AcaciaExpr, CMDLIST_T

class Result(NamedTuple):
    """Return value of a binary function implementation."""
    value: "AcaciaExpr"
    commands: "CMDLIST_T"

def commands(cmds: "CMDLIST_T") -> Result:
    return Result(objects.NoneLiteral(), cmds)

def literal(value: Union[bool, int, str, float, None]) -> "AcaciaExpr":
    if isinstance(value, bool):  # `bool` in front of `int`
        return objects.BoolLiteral(value)
    elif isinstance(value, int):
        return objects.IntLiteral(value)
    elif isinstance(value, str):
        return objects.String(value)
    elif isinstance(value, float):
        return objects.Float(value)
    elif value is None:
        return objects.NoneLiteral()
    unreachable(f"unexpected value {value!r}")



"""print - String formatting and printing module."""

from typing import List, Optional, Tuple, TYPE_CHECKING
from string import digits as DIGITS

from acaciamc.objects import *
from acaciamc.mccmdgen.expr import *
from acaciamc.mccmdgen.datatype import DefaultDataType
from acaciamc.mccmdgen.ctexpr import CTDataType
from acaciamc.mccmdgen.utils import InvalidOpError
from acaciamc.error import *
from acaciamc.tools import axe, resultlib
from acaciamc.constants import COLORS, COLORS_NEW
from acaciamc.localization import localize
import acaciamc.mccmdgen.cmds as cmds

if TYPE_CHECKING:
    from acaciamc.compiler import Compiler
    from acaciamc.mccmdgen.mcselector import MCSelector

class ArgFString(axe.Multityped):
    """Accepts an fstring as argument. A string is also accepted and
    converted to an fstring.
    """
    def __init__(self):
        super().__init__((StringDataType, FStringDataType))

    def convert(self, origin: AcaciaExpr) -> "FString":
        origin = super().convert(origin)
        if isinstance(origin, String):
            fstr = FString([], [cmds.RawtextText(origin.value)])
        else:
            assert isinstance(origin, FString)
            fstr = origin
        return fstr

class _FStrError(Exception):
    def __str__(self):
        return self.args[0]

class _FStrParser:
    def __init__(self, pattern: str, args, keywords, compiler: "Compiler"):
        """Parse an fstring with `pattern` and `args` and `keywords`
        as formatted expressions."""
        self.pattern = pattern
        self.ptr = 0
        self.args = args
        self.keywords = keywords
        self.compiler = compiler
        self.dependencies = []
        self.rawtext = cmds.Rawtext()

    def next_char(self) -> Optional[str]:
        """Move to next char."""
        if self.ptr >= len(self.pattern):
            return None
        c = self.pattern[self.ptr]
        self.ptr += 1
        return c

    def add_text(self, text: str):
        self.rawtext.append(cmds.RawtextText(text))

    def add_expr(self, expr: AcaciaExpr):
        """Format an expression to string."""
        try:
            components, deps = expr.rawtextify(self.compiler)
        except InvalidOpError:
            raise _FStrError(
                localize("modules.print.fstr.addexpr.error")
                % expr.data_type
            )
        else:
            self.rawtext.extend(components)
            self.dependencies.extend(deps)

    def expr_from_id(self, name: str) -> AcaciaExpr:
        """Get the expr from an format
        e.g. "0" -> args[0]; "key" -> keywords["key"]
        """
        if all(c in DIGITS for c in name):
            index = int(name)
            try:
                expr = self.args[index]
            except IndexError:
                raise _FStrError(
                    localize("modules.print.fstr.exprfromid.outofrange")
                    % index
                )
        elif name in self.keywords:
            expr = self.keywords[name]
        else:
            raise _FStrError(
                localize('modules.print.fstr.exprfromid.invalid') % name
            )
        return expr

    def parse(self):
        char = ''
        while True:
            char = self.next_char()
            if char is None:
                break
            # Normal char
            if char != '%':
                self.add_text(char)
                continue
            # % format
            second = self.next_char()
            if second == '%' or second is None:
                # '%%' escape -> '%' OR '%' at the end of string
                self.add_text('%')
                continue
            # 1. Parse which expression is being used here
            if second == '{':
                # read until }
                expr_str = []
                expr_char = self.next_char()
                while expr_char != '}':
                    if expr_char is None:
                        raise _FStrError(
                            localize("modules.print.fstr.parse.unclosedbrace")
                        )
                    expr_str.append(expr_char)
                    expr_char = self.next_char()
                # expr is integer or an identifier
                expr = self.expr_from_id(''.join(expr_str))
            elif second in DIGITS:
                # %1 is the alias to %{1}
                expr = self.expr_from_id(second)
            else:
                # can't be understood, just use raw text
                self.add_text(char + second)
                continue
            # 3. Handle the `expr` we got
            self.add_expr(expr)
        return self.dependencies, self.rawtext

class FStringDataType(DefaultDataType):
    name = 'fstring'

ctdt_fstring = CTDataType("fstring")

class FString(ConstExprCombined):
    """A formatted string in JSON format."""
    cdata_type = ctdt_fstring

    def __init__(self, dependencies: List[str],
                 components: List[cmds.RawtextComponent]):
        """
        dependencies: commands to run before json rawtext is used
        components: rawtext components
        """
        super().__init__(FStringDataType())
        self.dependencies = dependencies
        self.rawtext = cmds.Rawtext(components)

    def rawtextify(self, compiler):
        return self.rawtext, self.dependencies

    def copy(self):
        return FString(self.dependencies.copy(), self.rawtext.copy())

    def cadd(self, other):
        """Concatenate formatted strings."""
        res = self.copy()
        if isinstance(other, String):
            res.rawtext.append(cmds.RawtextText(other.value))
        elif isinstance(other, FString):
            res.rawtext.extend(other.rawtext)
        else:
            raise InvalidOpError
        return res

    def cradd(self, other):
        # Only normal strings can be `other`, since if they are fstrings
        # `cadd` would have handled it.
        res = self.copy()
        if isinstance(other, String):
            res.rawtext.insert(0, cmds.RawtextText(other.value))
        else:
            raise InvalidOpError
        return res

### Functions ###

## For creating strings

@axe.chop
@axe.arg("_pattern", axe.LiteralString(), rename="pattern")
@axe.star_arg("args", axe.AnyRT())
@axe.kwds("kwds", axe.AnyRT())
def _format(compiler, pattern: str, args, kwds):
    """
    format(_pattern: str, *args, **kwargs)

    Formatted string type.
    So far, `args` and `kwargs` may be int, bool, str and fstring
    NOTE Booleans are formatted as "0" and "1".
    in _pattern:
     "%%" -> character "%"
     "%{" integer "}" -> args[integer]
     "%" one-digit-number -> alias to `"%{" one-digit-number "}"`
     "%{" id "}" -> kwargs[id] (id is an valid Acacia identifier)
    Examples:
     format("Val1: %0; Val2: %{value}; Name: %{1}", val1, x, value=val2)
    """
    try:
        dependencies, rawtext = \
            _FStrParser(pattern, args, kwds, compiler).parse()
    except _FStrError as err:
        raise Error(ErrorType.ANY, message=str(err))
    return FString(dependencies, rawtext)

@axe.chop
@axe.arg("key", axe.LiteralString())
def _translate(compiler, key: str):
    """
    Return an fstring that uses the given localization key.
    Example: format("Give me %0!", translate("item.diamond.name"))
    """
    return FString([], [cmds.RawtextTranslate(key)])

class _FontComponent(cmds.RawtextText):
    """A utility component used by `with_font`."""

    def __init__(self, text: str, label: str):
        super().__init__(text)
        self.label = label

COLOR_DEFAULT = "default"

@axe.chop
@axe.arg("text", ArgFString())
@axe.arg("color", axe.LiteralStringEnum(*COLORS, *COLORS_NEW, COLOR_DEFAULT),
         default=COLOR_DEFAULT)
@axe.arg("bold", axe.LiteralBool(), default=False)
@axe.arg("italic", axe.LiteralBool(), default=False)
@axe.arg("obfuscated", axe.LiteralBool(), default=False)
def _with_font(compiler: "Compiler", text: FString, color: str,
               bold: bool, italic: bool, obfuscated: bool):
    """
    with_font(
        text: str | fstring,
        color: str = "default",
        bold: bool-literal = False,
        italic: bool-literal = False,
        obfuscated: bool-literal = False
    ) -> fstring
    Return an fstring that has the given color and font when displayed.
    When uses of this function are embedded, for example:
        with_font(
            format("Hello %0!", with_font("world", "red", bold=True)),
            color="green"
        )
    The inner `with_font` has higher priority, so the result is:
        "Hello " (green) + "world" (red, bold) + "!" (green)
    When this function is used, it is not recommended to use formatting
    code (\u00A7, section sign; \# escape in Acacia) manually.
    """
    if color in COLORS_NEW and compiler.cfg.mc_version < (1, 19, 80):
        raise axe.ArgumentError(
            "color", localize("modules.print.withfont.onlyhighversion") % color
        )
    res = text.copy()
    fmts = []
    if color != COLOR_DEFAULT:
        if color in COLORS_NEW:
            color_str = COLORS_NEW[color]
        else:
            color_str = COLORS[color]
        fmts.append("\xA7" + color_str)
    if bold:
        fmts.append("\xA7l")
    if italic:
        fmts.append("\xA7o")
    if obfuscated:
        fmts.append("\xA7k")
    scopes: List[Tuple[int, int]] = []  # [start, end)
    start_i = 0
    start_levels = 0
    length = len(res.rawtext)
    for i, component in enumerate(res.rawtext):
        if isinstance(component, _FontComponent):
            if component.label == "start":
                start_levels += 1
                if start_levels == 1 and i > start_i:
                    scopes.append((start_i, i))
                    start_i = -1
            if component.label == "end":
                start_levels -= 1
                if start_levels == 0 and i != length - 1:
                    start_i = i + 1
    if start_levels:
        raise ValueError("Unclosed font scopes")
    if start_i != -1:
        scopes.append((start_i, length))
    scopes.reverse()
    fmt_code = "".join(fmts)
    for start_i, end_i in scopes:
        res.rawtext.insert(end_i, _FontComponent("\xA7r", "end"))
        res.rawtext.insert(start_i, _FontComponent(fmt_code, "start"))
    return res

## For printing

@axe.chop
@axe.arg("text", ArgFString())
@axe.arg("target", axe.PlayerSelector(), default=None)
def _tell(compiler, text: FString, target: Optional["MCSelector"]):
    """tell(text: str | fstring, target: PlayerSelector = <All players>)
    Tell the `target` the `text` using /tellraw.
    """
    commands = []
    if target is None:
        target_str = "@a"
    else:
        target_str = target.to_str()
    commands.extend(text.dependencies)
    commands.append(cmds.RawtextOutput(f"tellraw {target_str}", text.rawtext))
    return resultlib.commands(commands)

# Title modes
_TITLE = 'title'
_SUBTITLE = 'subtitle'
_ACTIONBAR = 'actionbar'
# Default configurations
_FADE_IN = 10
_STAY_TIME = 70
_FADE_OUT = 20
_DEF_TITLE_CONFIG = (_FADE_IN, _STAY_TIME, _FADE_OUT)

@axe.chop
@axe.arg("text", ArgFString())
@axe.arg("target", axe.PlayerSelector(), default=None)
@axe.arg("mode", axe.LiteralStringEnum(_TITLE, _SUBTITLE, _ACTIONBAR),
         default=_TITLE)
@axe.arg("fade_in", axe.LiteralInt(), default=_FADE_IN)
@axe.arg("stay_time", axe.LiteralInt(), default=_STAY_TIME)
@axe.arg("fade_out", axe.LiteralInt(), default=_FADE_OUT)
def _title(compiler, text: FString, target: Optional["MCSelector"], mode: str,
           fade_in: int, stay_time: int, fade_out: int):
    """title(
        text: str | fstring,
        target: PlayerSelector = <All players>,
        mode: str = TITLE,
        fade_in: int-literal = 10,
        stay_time: int-literal = 70,
        fade_out: int-literal = 20
    )
    Use /titleraw for showing `text`.
    `fade_in`, `stay_time` and `fade_out` are in ticks.
    """
    commands = []
    if target is None:
        target_str = "@a"
    else:
        target_str = target.to_str()
    # Start
    ## set config
    conf = (fade_in, stay_time, fade_out)
    if conf != _DEF_TITLE_CONFIG:
        # only set config when it's not the default one
        commands.append(cmds.TitlerawTimes(target_str, *conf))
    ## titleraw
    commands.extend(text.dependencies)
    commands.append(
        cmds.RawtextOutput(f'titleraw {target_str} {mode}', text.rawtext)
    )
    ## reset config
    if conf != _DEF_TITLE_CONFIG:
        # only reset when config is not the default one
        commands.append(cmds.TitlerawResetTimes(target_str))
    ## return
    return resultlib.commands(commands)

@axe.chop
@axe.arg("target", axe.PlayerSelector(), default=None)
def _title_clear(compiler, target: Optional["MCSelector"]):
    """title_clear(target: PlayerSelector = <All players>)
    Clear `target`'s title text.
    """
    if target is None:
        target_str = "@a"
    else:
        target_str = target.to_str()
    return resultlib.commands([cmds.TitlerawClear(target_str)])

def acacia_build(compiler: "Compiler"):
    return {
        'format': BinaryCTFunction(_format),
        'translate': BinaryCTFunction(_translate),
        'with_font': BinaryCTFunction(_with_font),
        'tell': BinaryFunction(_tell),
        'title': BinaryFunction(_title),
        'TITLE': String(_TITLE),
        'SUBTITLE': String(_SUBTITLE),
        'ACTIONBAR': String(_ACTIONBAR),
        'title_clear': BinaryFunction(_title_clear)
    }



"""Builtin string."""

__all__ = ['StringDataType', 'String']

from acaciamc.mccmdgen.expr import *
from acaciamc.mccmdgen.datatype import DefaultDataType
from acaciamc.mccmdgen.ctexpr import CTDataType
from acaciamc.mccmdgen.utils import InvalidOpError

class StringDataType(DefaultDataType):
    name = 'str'

ctdt_string = CTDataType("str")

class String(ConstExprCombined):
    cdata_type = ctdt_string

    def __init__(self, value: str):
        super().__init__(StringDataType())
        self.value = value

    def chash(self):
        return self.value

    def cstringify(self) -> str:
        return self.value

    def cadd(self, other):
        """Adding strings will connect them."""
        if isinstance(other, String):
            return String(self.value + other.value)
        raise InvalidOpError
